import json
import logging
import os
from typing import Any, Optional

from r2r.core import Prompt, PromptProvider

logger = logging.getLogger(__name__)


class R2RPromptProvider(PromptProvider):
    def __init__(self, file_path: Optional[str] = None):
        self.prompts: dict[str, Prompt] = {}
        self._load_prompts_from_jsonl(file_path=file_path)

    def _load_prompts_from_jsonl(self, file_path: Optional[str] = None):
        if not file_path:
            file_path = os.path.join(
                os.path.dirname(__file__), "defaults.jsonl"
            )
        try:
            with open(file_path, "r") as file:
                for line in file:
                    if line.strip():
                        data = json.loads(line)
                        self.add_prompt(
                            data["name"],
                            data["template"],
                            data.get("input_types", {}),
                        )
        except json.JSONDecodeError as e:
            error_msg = f"Error loading prompts from JSONL file: {e}"
            logger.error(error_msg)
            raise ValueError(error_msg)

    def add_prompt(
        self, name: str, template: str, input_types: dict[str, str]
    ) -> None:
        if name in self.prompts:
            raise ValueError(f"Prompt '{name}' already exists.")
        self.prompts[name] = Prompt(
            name=name, template=template, input_types=input_types
        )

    def get_prompt(
        self, prompt_name: str, inputs: Optional[dict[str, Any]] = None
    ) -> str:
        if prompt_name not in self.prompts:
            raise ValueError(f"Prompt '{prompt_name}' not found.")
        prompt = self.prompts[prompt_name]
        if inputs is None:
            return prompt.template
        return prompt.format_prompt(inputs)

    def update_prompt(
        self,
        name: str,
        template: Optional[str] = None,
        input_types: Optional[dict[str, str]] = None,
    ) -> None:
        if name not in self.prompts:
            raise ValueError(f"Prompt '{name}' not found.")
        if template:
            self.prompts[name].template = template
        if input_types:
            self.prompts[name].input_types = input_types

    def get_all_prompts(self) -> dict[str, Prompt]:
        return self.prompts



import logging
import random

from r2r.core import EmbeddingConfig, EmbeddingProvider, SearchResult

logger = logging.getLogger(__name__)


class DummyEmbeddingProvider(EmbeddingProvider):
    def __init__(self, config: EmbeddingConfig):
        logger.info(
            "Initializing `DummyEmbeddingProvider` to provide embeddings."
        )
        super().__init__(config)
        provider = config.provider
        if not provider:
            raise ValueError(
                "Must set provider in order to initialize SentenceTransformerEmbeddingProvider."
            )

        self.base_dimension = config.base_dimension

        if not self.base_dimension:
            raise ValueError(
                "Must set base_dimension in order to initialize DummyEmbeddingProvider."
            )

        if config.rerank_model:
            raise ValueError(
                "DummyEmbeddingProvider does not support separate reranking."
            )

    def get_embedding(
        self,
        text: str,
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.BASE,
    ) -> list[float]:
        if stage != EmbeddingProvider.PipeStage.BASE:
            raise ValueError(
                "DummyEmbeddingProvider only supports search stage."
            )
        return [random.random() for _ in range(self.base_dimension)]

    def get_embeddings(
        self,
        texts: list[str],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.BASE,
    ) -> list[list[float]]:
        if stage != EmbeddingProvider.PipeStage.BASE:
            raise ValueError(
                "DummyEmbeddingProvider only supports search stage."
            )

        return [
            [random.random() for _ in range(self.base_dimension)]
            for text in texts
        ]

    # TODO: This should be the default (base class) implementation instead of 'pass'.
    def rerank(
        self,
        transformed_query: str,
        texts: list[SearchResult],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.RERANK,
        limit: int = 10,
    ):
        return texts[:limit]

    def tokenize_string(
        self,
        text: str,
        model: str,
        stage: EmbeddingProvider.PipeStage,
    ) -> list[int]:
        """Tokenizes the input string."""
        return [0]



import uuid
from typing import Any, Dict


class BaseTelemetryEvent:
    def __init__(self, event_type: str, properties: Dict[str, Any]):
        self.event_type = event_type
        self.properties = properties
        self.event_id = str(uuid.uuid4())


class DailyActiveUserEvent(BaseTelemetryEvent):
    def __init__(self, user_id: str):
        super().__init__("DailyActiveUser", {"user_id": user_id})


class FeatureUsageEvent(BaseTelemetryEvent):
    def __init__(self, user_id: str, feature: str):
        super().__init__(
            "FeatureUsage", {"user_id": user_id, "feature": feature}
        )


class ErrorEvent(BaseTelemetryEvent):
    def __init__(self, user_id: str, endpoint: str, error_message: str):
        super().__init__(
            "Error",
            {
                "user_id": user_id,
                "endpoint": endpoint,
                "error_message": error_message,
            },
        )


class RequestLatencyEvent(BaseTelemetryEvent):
    def __init__(self, endpoint: str, latency: float):
        super().__init__(
            "RequestLatency", {"endpoint": endpoint, "latency": latency}
        )


class GeographicDistributionEvent(BaseTelemetryEvent):
    def __init__(self, user_id: str, country: str):
        super().__init__(
            "GeographicDistribution", {"user_id": user_id, "country": country}
        )


class SessionDurationEvent(BaseTelemetryEvent):
    def __init__(self, user_id: str, duration: float):
        super().__init__(
            "SessionDuration", {"user_id": user_id, "duration": duration}
        )


class UserPathEvent(BaseTelemetryEvent):
    def __init__(self, user_id: str, path: str):
        super().__init__("UserPath", {"user_id": user_id, "path": path})



import logging
import random
import uuid
from typing import Any, AsyncGenerator, Optional

from pydantic import BaseModel

from r2r import (
    AsyncState,
    EvalProvider,
    GenerationConfig,
    LLMChatCompletion,
    LoggableAsyncPipe,
    PipeType,
)

logger = logging.getLogger(__name__)


class R2REvalPipe(LoggableAsyncPipe):
    class EvalPayload(BaseModel):
        query: str
        context: str
        completion: str

    class Input(LoggableAsyncPipe.Input):
        message: AsyncGenerator["R2REvalPipe.EvalPayload", None]

    def __init__(
        self,
        eval_provider: EvalProvider,
        type: PipeType = PipeType.EVAL,
        config: Optional[LoggableAsyncPipe.PipeConfig] = None,
        *args,
        **kwargs,
    ):
        self.eval_provider = eval_provider
        super().__init__(
            type=type,
            config=config
            or LoggableAsyncPipe.PipeConfig(name="default_eval_pipe"),
            *args,
            **kwargs,
        )

    async def _run_logic(
        self,
        input: Input,
        state: AsyncState,
        run_id: uuid.UUID,
        eval_generation_config: GenerationConfig,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[LLMChatCompletion, None]:
        async for item in input.message:
            yield self.eval_provider.evaluate(
                item.query,
                item.context,
                item.completion,
                eval_generation_config,
            )



import logging
import os

from openai import AsyncOpenAI, AuthenticationError, OpenAI

from r2r.core import EmbeddingConfig, EmbeddingProvider, SearchResult

logger = logging.getLogger(__name__)


class OpenAIEmbeddingProvider(EmbeddingProvider):
    MODEL_TO_TOKENIZER = {
        "text-embedding-ada-002": "cl100k_base",
        "text-embedding-3-small": "cl100k_base",
        "text-embedding-3-large": "cl100k_base",
    }
    MODEL_TO_DIMENSIONS = {
        "text-embedding-ada-002": [1536],
        "text-embedding-3-small": [512, 1536],
        "text-embedding-3-large": [256, 1024, 3072],
    }

    def __init__(self, config: EmbeddingConfig):
        logger.info(
            "Initializing `OpenAIEmbeddingProvider` to provide embeddings."
        )
        super().__init__(config)
        provider = config.provider
        if not provider:
            raise ValueError(
                "Must set provider in order to initialize OpenAIEmbeddingProvider."
            )

        if provider != "openai":
            raise ValueError(
                "OpenAIEmbeddingProvider must be initialized with provider `openai`."
            )
        if not os.getenv("OPENAI_API_KEY"):
            raise ValueError(
                "Must set OPENAI_API_KEY in order to initialize OpenAIEmbeddingProvider."
            )
        self.client = OpenAI()
        self.async_client = AsyncOpenAI()

        if config.rerank_model:
            raise ValueError(
                "OpenAIEmbeddingProvider does not support separate reranking."
            )
        self.base_model = config.base_model
        self.base_dimension = config.base_dimension

        if self.base_model not in OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER:
            raise ValueError(
                f"OpenAI embedding model {self.base_model} not supported."
            )
        if (
            self.base_dimension
            and self.base_dimension
            not in OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[self.base_model]
        ):
            raise ValueError(
                f"Dimensions {self.dimension} for {self.base_model} are not supported"
            )

        if not self.base_model or not self.base_dimension:
            raise ValueError(
                "Must set base_model and base_dimension in order to initialize OpenAIEmbeddingProvider."
            )

        if config.rerank_model:
            raise ValueError(
                "OpenAIEmbeddingProvider does not support separate reranking."
            )

    def get_embedding(
        self,
        text: str,
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.BASE,
    ) -> list[float]:
        if stage != EmbeddingProvider.PipeStage.BASE:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        try:
            return (
                self.client.embeddings.create(
                    input=[text],
                    model=self.base_model,
                    dimensions=self.base_dimension
                    or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[
                        self.base_model
                    ][-1],
                )
                .data[0]
                .embedding
            )
        except AuthenticationError as e:
            raise ValueError(
                "Invalid OpenAI API key provided. Please check your OPENAI_API_KEY environment variable."
            ) from e

    async def async_get_embedding(
        self,
        text: str,
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.BASE,
    ) -> list[float]:
        if stage != EmbeddingProvider.PipeStage.BASE:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        try:
            response = await self.async_client.embeddings.create(
                input=[text],
                model=self.base_model,
                dimensions=self.base_dimension
                or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[
                    self.base_model
                ][-1],
            )
            return response.data[0].embedding
        except AuthenticationError as e:
            raise ValueError(
                "Invalid OpenAI API key provided. Please check your OPENAI_API_KEY environment variable."
            ) from e

    def get_embeddings(
        self,
        texts: list[str],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.BASE,
    ) -> list[list[float]]:
        if stage != EmbeddingProvider.PipeStage.BASE:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        try:
            return [
                ele.embedding
                for ele in self.client.embeddings.create(
                    input=texts,
                    model=self.base_model,
                    dimensions=self.base_dimension
                    or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[
                        self.base_model
                    ][-1],
                ).data
            ]
        except AuthenticationError as e:
            raise ValueError(
                "Invalid OpenAI API key provided. Please check your OPENAI_API_KEY environment variable."
            ) from e

    async def async_get_embeddings(
        self,
        texts: list[str],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.BASE,
    ) -> list[list[float]]:
        if stage != EmbeddingProvider.PipeStage.BASE:
            raise ValueError(
                "OpenAIEmbeddingProvider only supports search stage."
            )

        try:
            response = await self.async_client.embeddings.create(
                input=texts,
                model=self.base_model,
                dimensions=self.base_dimension
                or OpenAIEmbeddingProvider.MODEL_TO_DIMENSIONS[
                    self.base_model
                ][-1],
            )
            return [ele.embedding for ele in response.data]
        except AuthenticationError as e:
            raise ValueError(
                "Invalid OpenAI API key provided. Please check your OPENAI_API_KEY environment variable."
            ) from e

    def rerank(
        self,
        query: str,
        results: list[SearchResult],
        stage: EmbeddingProvider.PipeStage = EmbeddingProvider.PipeStage.RERANK,
        limit: int = 10,
    ):
        return results[:limit]

    def tokenize_string(self, text: str, model: str) -> list[int]:
        try:
            import tiktoken
        except ImportError:
            raise ValueError(
                "Must download tiktoken library to run `tokenize_string`."
            )
        # tiktoken encoding -
        # cl100k_base -	gpt-4, gpt-3.5-turbo, text-embedding-ada-002, text-embedding-3-small, text-embedding-3-large
        if model not in OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER:
            raise ValueError(f"OpenAI embedding model {model} not supported.")
        encoding = tiktoken.get_encoding(
            OpenAIEmbeddingProvider.MODEL_TO_TOKENIZER[model]
        )
        return encoding.encode(text)



import logging
from abc import ABC, abstractmethod
from typing import Optional, Union

from ..abstractions.document import DocumentInfo
from ..abstractions.search import SearchResult
from ..abstractions.vector import VectorEntry
from .base_provider import Provider, ProviderConfig

logger = logging.getLogger(__name__)


class VectorDBConfig(ProviderConfig):
    provider: str
    collection_name: str

    def __post_init__(self):
        self.validate()
        # Capture additional fields
        for key, value in self.extra_fields.items():
            setattr(self, key, value)

    def validate(self) -> None:
        if self.provider not in self.supported_providers:
            raise ValueError(f"Provider '{self.provider}' is not supported.")

    @property
    def supported_providers(self) -> list[str]:
        return ["local", "pgvector"]


class VectorDBProvider(Provider, ABC):
    def __init__(self, config: VectorDBConfig):
        if not isinstance(config, VectorDBConfig):
            raise ValueError(
                "VectorDBProvider must be initialized with a `VectorDBConfig`."
            )
        logger.info(f"Initializing VectorDBProvider with config {config}.")
        super().__init__(config)

    @abstractmethod
    def initialize_collection(self, dimension: int) -> None:
        pass

    @abstractmethod
    def copy(self, entry: VectorEntry, commit: bool = True) -> None:
        pass

    @abstractmethod
    def upsert(self, entry: VectorEntry, commit: bool = True) -> None:
        pass

    @abstractmethod
    def search(
        self,
        query_vector: list[float],
        filters: dict[str, Union[bool, int, str]] = {},
        limit: int = 10,
        *args,
        **kwargs,
    ) -> list[SearchResult]:
        pass

    @abstractmethod
    def hybrid_search(
        self,
        query_text: str,
        query_vector: list[float],
        limit: int = 10,
        filters: Optional[dict[str, Union[bool, int, str]]] = None,
        # Hybrid search parameters
        full_text_weight: float = 1.0,
        semantic_weight: float = 1.0,
        rrf_k: int = 20,  # typical value is ~2x the number of results you want
        *args,
        **kwargs,
    ) -> list[SearchResult]:
        pass

    @abstractmethod
    def create_index(self, index_type, column_name, index_options):
        pass

    def upsert_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        for entry in entries:
            self.upsert(entry, commit=commit)

    def copy_entries(
        self, entries: list[VectorEntry], commit: bool = True
    ) -> None:
        for entry in entries:
            self.copy(entry, commit=commit)

    @abstractmethod
    def delete_by_metadata(
        self,
        metadata_fields: list[str],
        metadata_values: list[Union[bool, int, str]],
    ) -> list[str]:
        if len(metadata_fields) != len(metadata_values):
            raise ValueError(
                "The number of metadata fields and values must be equal."
            )
        pass

    @abstractmethod
    def get_metadatas(
        self,
        metadata_fields: list[str],
        filter_field: Optional[str] = None,
        filter_value: Optional[str] = None,
    ) -> list[str]:
        pass

    @abstractmethod
    def upsert_documents_info(self, document_infs: list[DocumentInfo]) -> None:
        pass

    @abstractmethod
    def get_documents_info(
        self,
        filter_document_ids: Optional[list[str]] = None,
        filter_user_ids: Optional[list[str]] = None,
    ) -> list[DocumentInfo]:
        pass

    @abstractmethod
    def get_document_chunks(self, document_id: str) -> list[dict]:
        pass

    @abstractmethod
    def delete_documents_info(self, document_ids: list[str]) -> dict:
        pass

    @abstractmethod
    def get_users_stats(self, user_ids: Optional[list[str]] = None) -> dict:
        pass



"""Abstractions for search functionality."""

import uuid
from typing import Any, Optional

from pydantic import BaseModel


class SearchRequest(BaseModel):
    """Request for a search operation."""

    query: str
    limit: int
    filters: Optional[dict[str, Any]] = None


class SearchResult(BaseModel):
    """Result of a search operation."""

    id: uuid.UUID
    score: float
    metadata: dict[str, Any]

    def __str__(self) -> str:
        return f"SearchResult(id={self.id}, score={self.score}, metadata={self.metadata})"

    def __repr__(self) -> str:
        return f"SearchResult(id={self.id}, score={self.score}, metadata={self.metadata})"

    def dict(self) -> dict:
        return {
            "id": self.id,
            "score": self.score,
            "metadata": self.metadata,
        }



import logging
import uuid
from typing import Any, Optional

from r2r.core import (
    AsyncState,
    GenerationConfig,
    KGProvider,
    KVLoggingSingleton,
    LLMProvider,
    PipeType,
    PromptProvider,
)

from .abstractions.generator_pipe import GeneratorPipe

logger = logging.getLogger(__name__)


prompt = """**System Message:**

You are an AI assistant capable of generating Cypher queries to interact with a Neo4j knowledge graph. The knowledge graph contains information about organizations, people, locations, and their relationships, such as founders of companies, locations of companies, and products associated with companies.

**Instructions:**

When a user asks a question, you will generate a Cypher query to retrieve the relevant information from the Neo4j knowledge graph. Later, you will be given a schema which specifies the available relationships to help you construct the query. First, review the examples provided to understand the expected format of the queries.

### Example(s) - User Questions and Cypher Queries for an Academic Knowledge Graph

**User Question:**
"List all courses available in the computer science department."

**Generated Cypher Query:**
```cypher
MATCH (c:COURSE)-[:OFFERED_BY]->(d:DEPARTMENT)
WHERE d.name CONTAINS 'Computer Science'
RETURN c.id AS Course, d.name AS Department
ORDER BY c.id;
```

**User Question:**
"Retrieve all courses taught by professors who have published research on natural language processing."

**Generated Cypher Query:**
```cypher
MATCH (pr:PERSON)-[:PUBLISHED]->(p:PAPER)
MATCH (p)-[:TOPIC]->(t:TOPIC)
WHERE t.name CONTAINS 'Natural Language Processing'
MATCH (c:COURSE)-[:TAUGHT_BY]->(pr)
RETURN DISTINCT c.id AS Course, pr.name AS Professor, t.name AS Topic
ORDER BY c.id;
```


### Example(s) - User Questions and Cypher Queries for an Historical Events and Figures

**User Question:**
"List all battles that occurred in the 19th century and the generals who participated in them."

**Generated Cypher Query:**
```cypher
MATCH (b:EVENT)-[:HAPPENED_AT]->(d:DATE)
WHERE d.year >= 1800 AND d.year < 1900 AND b.type CONTAINS 'Battle'
MATCH (g:PERSON)-[:PARTICIPATED_IN]->(b)
RETURN b.name AS Battle, d.year AS Year, g.name AS General
ORDER BY d.year, b.name, g.name;
```

**User Question:**
"Find all treaties signed in Paris and the countries involved."


**Generated Cypher Query:**
```cypher
MATCH (t:EVENT)-[:HAPPENED_AT]->(l:LOCATION)
WHERE l.name CONTAINS 'Paris' AND t.type CONTAINS 'Treaty'
MATCH (c:ORGANIZATION)-[:SIGNED]->(t)
RETURN t.name AS Treaty, l.name AS Location, c.name AS Country
ORDER BY t.name, c.name;
```


Now, you will be provided with a schema for the entities and relationships in the Neo4j knowledge graph. Use this schema to construct Cypher queries based on user questions.

- **Entities:**
  - `ORGANIZATION` (e.g.: `COMPANY`, `SCHOOL`, `NON-PROFIT`, `OTHER`)
  - `LOCATION` (e.g.: `CITY`, `STATE`, `COUNTRY`, `OTHER`)
  - `PERSON`
  - `POSITION`
  - `DATE` (e.g.: `YEAR`, `MONTH`, `DAY`, `BATCH`, `OTHER`)
  - `QUANTITY`
  - `EVENT` (e.g.: `INCORPORATION`, `FUNDING_ROUND`, `ACQUISITION`, `LAUNCH`, `OTHER`)
  - `INDUSTRY`
  - `MEDIA` (e.g.: `EMAIL`, `WEBSITE`, `TWITTER`, `LINKEDIN`, `OTHER`)
  - `PRODUCT`

- **Relationships:**
  - `FOUNDED`
  - `WORKED_AT`
  - `EDUCATED_AT`
  - `RAISED`
  - `REVENUE`
  - `TEAM_SIZE`
  - `LOCATION`
  - `ACQUIRED_BY`
  - `ANNOUNCED`
  - `INDUSTRY`
  - `PRODUCT`
  - `FEATURES`
  - `USES`
  - `USED_BY`
  - `TECHNOLOGY`
  - `HAS`
  - `AS_OF`
  - `PARTICIPATED`
  - `ASSOCIATED`
  - `GROUP_PARTNER`
  - `ALIAS`

Use the referenced examples and schema to help you construct an appropriate Cypher query based on the following question:

**User Question:**
{question}

**Generated Cypher Query:**
"""


class R2RKGAgentPipe(GeneratorPipe):
    """
    Embeds and stores documents using a specified embedding model and database.
    """

    def __init__(
        self,
        kg_provider: KGProvider,
        llm_provider: LLMProvider,
        prompt_provider: PromptProvider,
        pipe_logger: Optional[KVLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        config: Optional[GeneratorPipe.PipeConfig] = None,
        *args,
        **kwargs,
    ):
        """
        Initializes the embedding pipe with necessary components and configurations.
        """
        super().__init__(
            llm_provider=llm_provider,
            prompt_provider=prompt_provider,
            type=type,
            config=config
            or GeneratorPipe.Config(
                name="kg_rag_pipe", task_prompt="kg_agent"
            ),
            pipe_logger=pipe_logger,
            *args,
            **kwargs,
        )
        self.kg_provider = kg_provider
        self.llm_provider = llm_provider
        self.prompt_provider = prompt_provider
        self.pipe_run_info = None

    async def _run_logic(
        self,
        input: GeneratorPipe.Input,
        state: AsyncState,
        run_id: uuid.UUID,
        rag_generation_config: GenerationConfig,
        *args: Any,
        **kwargs: Any,
    ):

        async for message in input.message:
            # TODO - Remove hard code
            formatted_prompt = self.prompt_provider.get_prompt(
                "kg_agent", {"input": message}
            )
            messages = self._get_message_payload(formatted_prompt)

            result = self.llm_provider.get_completion(
                messages=messages, generation_config=rag_generation_config
            )

            extraction = result.choices[0].message.content
            query = extraction.split("```cypher")[1].split("```")[0]
            yield self.kg_provider.structured_query(query)

    def _get_message_payload(self, message: str) -> dict:
        return [
            {
                "role": "system",
                "content": self.prompt_provider.get_prompt(
                    self.config.system_prompt,
                ),
            },
            {"role": "user", "content": message},
        ]



from typing import Optional

from pydantic import BaseModel

from r2r.core import (
    EmbeddingProvider,
    EvalPipeline,
    EvalProvider,
    IngestionPipeline,
    KGProvider,
    LLMProvider,
    LoggableAsyncPipe,
    PromptProvider,
    RAGPipeline,
    SearchPipeline,
    VectorDBProvider,
)


class R2RProviders(BaseModel):
    vector_db: Optional[VectorDBProvider]
    embedding: Optional[EmbeddingProvider]
    llm: Optional[LLMProvider]
    prompt: Optional[PromptProvider]
    eval: Optional[EvalProvider]
    kg: Optional[KGProvider]

    class Config:
        arbitrary_types_allowed = True


class R2RPipes(BaseModel):
    parsing_pipe: LoggableAsyncPipe
    embedding_pipe: LoggableAsyncPipe
    vector_storage_pipe: LoggableAsyncPipe
    search_pipe: LoggableAsyncPipe
    rag_pipe: LoggableAsyncPipe
    streaming_rag_pipe: LoggableAsyncPipe
    eval_pipe: LoggableAsyncPipe
    kg_pipe: LoggableAsyncPipe
    kg_storage_pipe: LoggableAsyncPipe

    class Config:
        arbitrary_types_allowed = True


class R2RPipelines(BaseModel):
    eval_pipeline: EvalPipeline
    ingestion_pipeline: IngestionPipeline
    search_pipeline: SearchPipeline
    rag_pipeline: RAGPipeline
    streaming_rag_pipeline: RAGPipeline

    class Config:
        arbitrary_types_allowed = True



from abc import ABC, abstractmethod, abstractproperty
from typing import Any, Optional, Type

from pydantic import BaseModel


class ProviderConfig(BaseModel, ABC):
    """A base provider configuration class"""

    extra_fields: dict[str, Any] = {}
    provider: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True
        ignore_extra = True

    @abstractmethod
    def validate(self) -> None:
        pass

    @classmethod
    def create(cls: Type["ProviderConfig"], **kwargs: Any) -> "ProviderConfig":
        base_args = cls.__fields__.keys()
        filtered_kwargs = {
            k: v if v != "None" else None
            for k, v in kwargs.items()
            if k in base_args
        }
        instance = cls(**filtered_kwargs)
        for k, v in kwargs.items():
            if k not in base_args:
                instance.extra_fields[k] = v
        return instance

    @abstractproperty
    @property
    def supported_providers(self) -> list[str]:
        """Define a list of supported providers."""
        pass


class Provider(ABC):
    """A base provider class to provide a common interface for all providers."""

    def __init__(self, config: Optional[ProviderConfig] = None):
        if config:
            config.validate()
        self.config = config



import uuid
from copy import copy
from typing import Any, AsyncGenerator, Optional

from r2r import (
    GenerationConfig,
    LoggableAsyncPipe,
    R2RPipeFactory,
    R2RQueryTransformPipe,
    SearchPipe,
    SearchResult,
)


class MultiSearchPipe(LoggableAsyncPipe):
    class PipeConfig(LoggableAsyncPipe.PipeConfig):
        name: str = "multi_search_pipe"

    def __init__(
        self,
        query_transform_pipe: R2RQueryTransformPipe,
        inner_search_pipe: SearchPipe,
        config: Optional[PipeConfig] = None,
        *args,
        **kwargs,
    ):
        self.query_transform_pipe = query_transform_pipe
        self.search_pipe = inner_search_pipe
        if (
            not query_transform_pipe.config.name
            == inner_search_pipe.config.name
        ):
            raise ValueError(
                "The query transform pipe and search pipe must have the same name."
            )
        if config and not config.name == query_transform_pipe.config.name:
            raise ValueError(
                "The pipe config name must match the query transform pipe name."
            )

        super().__init__(
            config=config
            or MultiSearchPipe.PipeConfig(
                name=query_transform_pipe.config.name
            ),
            *args,
            **kwargs,
        )

    async def _run_logic(
        self,
        input: Any,
        state: Any,
        run_id: uuid.UUID,
        query_transform_generation_config: Optional[GenerationConfig] = None,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[SearchResult, None]:
        query_transform_generation_config = (
            query_transform_generation_config
            or copy(kwargs.get("rag_generation_config", None))
            or GenerationConfig(model="gpt-4o")
        )
        query_transform_generation_config.stream = False

        query_generator = await self.query_transform_pipe.run(
            input,
            state,
            query_transform_generation_config=query_transform_generation_config,
            num_query_xf_outputs=3,
            *args,
            **kwargs,
        )

        async for search_result in await self.search_pipe.run(
            self.search_pipe.Input(message=query_generator),
            state,
            *args,
            **kwargs,
        ):
            yield search_result


class R2RPipeFactoryWithMultiSearch(R2RPipeFactory):
    QUERY_GENERATION_TEMPLATE: dict = (
        {  # TODO - Can we have stricter typing like so? `: {"template": str, "input_types": dict[str, str]} = {``
            "template": "### Instruction:\n\nGiven the following query that follows to write a double newline separated list of up to {num_outputs} queries meant to help answer the original query. \nDO NOT generate any single query which is likely to require information from multiple distinct documents, \nEACH single query will be used to carry out a cosine similarity semantic search over distinct indexed documents, such as varied medical documents. \nFOR EXAMPLE if asked `how do the key themes of Great Gatsby compare with 1984`, the two queries would be \n`What are the key themes of Great Gatsby?` and `What are the key themes of 1984?`.\nHere is the original user query to be transformed into answers:\n\n### Query:\n{message}\n\n### Response:\n",
            "input_types": {"num_outputs": "int", "message": "str"},
        }
    )

    def create_search_pipe(self, *args, **kwargs):
        """
        A factory method to create a search pipe.

        Overrides include
            multi_query_transform_pipe_override: R2RQueryTransformPipe
            multi_inner_search_pipe_override: SearchPipe
            search_task_template_override: {'template': str, 'input_types': dict[str, str]}
        """
        multi_search_config = MultiSearchPipe.PipeConfig()
        task_prompt_name = (
            kwargs.get("task_prompt_name")
            or f"{multi_search_config.name}_task_prompt"
        )

        # Initialize the new query transform pipe
        query_transform_pipe = kwargs.get(
            "multi_query_transform_pipe_override", None
        ) or R2RQueryTransformPipe(
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            config=R2RQueryTransformPipe.QueryTransformConfig(
                name=multi_search_config.name,
                task_prompt=task_prompt_name,
            ),
        )
        if kwargs.get("task_prompt_name") == None:
            # Add a prompt for transforming the user query
            self.providers.prompt.add_prompt(
                name=task_prompt_name,
                **(
                    kwargs.get("query_generation_template_override")
                    or self.QUERY_GENERATION_TEMPLATE
                ),
            )

        # Create search pipe override and pipes
        inner_search_pipe = kwargs.get(
            "multi_inner_search_pipe_override", None
        ) or super().create_search_pipe(*args, **kwargs)

        # TODO - modify `create_..._pipe` to allow naming the pipe
        inner_search_pipe.config.name = multi_search_config.name

        return MultiSearchPipe(
            query_transform_pipe=query_transform_pipe,
            inner_search_pipe=inner_search_pipe,
            config=multi_search_config,
        )



import asyncio
import json
import uuid
from typing import AsyncGenerator, Generator, Optional, Union

import httpx
import nest_asyncio
import requests

from r2r.core import DocumentType

nest_asyncio.apply()


def default_serializer(obj):
    if isinstance(obj, uuid.UUID):
        return str(obj)
    if isinstance(obj, DocumentType):
        return obj.value
    if isinstance(obj, bytes):
        raise TypeError("Bytes serialization is not yet supported.")
    raise TypeError(f"Type {type(obj)} not serializable.")


class R2RClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def update_prompt(
        self,
        name: str,
        template: Optional[str] = None,
        input_types: Optional[dict] = None,
    ) -> dict:
        url = f"{self.base_url}/update_prompt"
        data = {
            "name": name,
            "template": template,
            "input_types": input_types,
        }
        response = requests.post(
            url,
            data=json.dumps(data, default=default_serializer),
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()
        return response.json()

    def ingest_documents(self, documents: list[dict]) -> dict:
        url = f"{self.base_url}/ingest_documents"
        data = {"documents": documents}
        serialized_data = json.dumps(data, default=default_serializer)
        response = requests.post(
            url,
            data=serialized_data,
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()
        return response.json()

    def ingest_files(
        self,
        files: list[str],
        metadatas: Optional[list[dict]] = None,
        ids: Optional[list[str]] = None,
        user_ids: Optional[list[str]] = None,
    ) -> dict:
        url = f"{self.base_url}/ingest_files"
        files_to_upload = [
            ("files", (file, open(file, "rb"), "application/octet-stream"))
            for file in files
        ]
        data = {
            "metadatas": (
                None
                if metadatas is None
                else json.dumps(metadatas, default=default_serializer)
            ),
            "document_ids": (
                None
                if ids is None
                else json.dumps(ids, default=default_serializer)
            ),
            "user_ids": (
                None
                if user_ids is None
                else json.dumps(user_ids, default=default_serializer)
            ),
        }
        response = requests.post(url, files=files_to_upload, data=data)
        response.raise_for_status()
        return response.json()

    def update_documents(self, documents: list[dict]) -> dict:
        url = f"{self.base_url}/update_documents"
        data = {"documents": documents}
        serialized_data = json.dumps(data, default=default_serializer)
        response = requests.post(
            url,
            data=serialized_data,
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()
        return response.json()

    def update_files(
        self,
        files: list[str],
        ids: list[str],
        metadatas: Optional[list[dict]] = None,
    ) -> dict:
        url = f"{self.base_url}/update_files"
        files_to_upload = [
            ("files", (file, open(file, "rb"), "application/octet-stream"))
            for file in files
        ]
        data = {
            "metadatas": (
                None
                if metadatas is None
                else json.dumps(metadatas, default=default_serializer)
            ),
            "ids": json.dumps(ids, default=default_serializer),
        }
        response = requests.post(url, files=files_to_upload, data=data)
        response.raise_for_status()
        return response.json()

    def search(
        self,
        query: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
        do_hybrid_search: bool = False,
    ) -> dict:
        url = f"{self.base_url}/search"
        data = {
            "query": query,
            "search_filters": json.dumps(search_filters or {}),
            "search_limit": search_limit,
            "do_hybrid_search": do_hybrid_search,
        }
        response = requests.post(url, json=data)
        response.raise_for_status()
        return response.json()

    def rag(
        self,
        message: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
        rag_generation_config: Optional[dict] = None,
        streaming: bool = False,
    ) -> Union[dict, Generator[str, None, None]]:
        if streaming:
            return self._stream_rag_sync(
                message=message,
                search_filters=search_filters,
                search_limit=search_limit,
                rag_generation_config=rag_generation_config,
            )
        else:
            try:
                url = f"{self.base_url}/rag"
                data = {
                    "message": message,
                    "search_filters": (
                        json.dumps(search_filters) if search_filters else None
                    ),
                    "search_limit": search_limit,
                    "rag_generation_config": (
                        json.dumps(rag_generation_config)
                        if rag_generation_config
                        else None
                    ),
                    "streaming": streaming,
                }

                response = requests.post(url, json=data)
                response.raise_for_status()
                return response.json()
            except requests.exceptions.RequestException as e:
                raise e

    async def _stream_rag(
        self,
        message: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
        rag_generation_config: Optional[dict] = None,
    ) -> AsyncGenerator[str, None]:
        url = f"{self.base_url}/rag"
        data = {
            "message": message,
            "search_filters": (
                json.dumps(search_filters) if search_filters else None
            ),
            "search_limit": search_limit,
            "rag_generation_config": (
                json.dumps(rag_generation_config)
                if rag_generation_config
                else None
            ),
            "streaming": True,
        }
        async with httpx.AsyncClient() as client:
            async with client.stream("POST", url, json=data) as response:
                response.raise_for_status()
                async for chunk in response.aiter_text():
                    yield chunk

    def _stream_rag_sync(
        self,
        message: str,
        search_filters: Optional[dict] = None,
        search_limit: int = 10,
        rag_generation_config: Optional[dict] = None,
    ) -> Generator[str, None, None]:
        async def run_async_generator():
            async for chunk in self._stream_rag(
                message=message,
                search_filters=search_filters,
                search_limit=search_limit,
                rag_generation_config=rag_generation_config,
            ):
                yield chunk

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # Run the async generator and yield each chunk synchronously
        async_gen = run_async_generator()

        try:
            while True:
                # Fetch the next chunk from the async generator
                chunk = loop.run_until_complete(async_gen.__anext__())
                yield chunk
        except StopAsyncIteration:
            pass
        finally:
            loop.close()

    def delete(
        self, keys: list[str], values: list[Union[bool, int, str]]
    ) -> dict:
        url = f"{self.base_url}/delete"
        data = {"keys": keys, "values": values}
        response = requests.request("DELETE", url, json=data)
        response.raise_for_status()
        return response.json()

    def logs(self, log_type_filter: Optional[str] = None) -> dict:
        url = f"{self.base_url}/logs"
        params = {}
        if log_type_filter:
            params["log_type_filter"] = log_type_filter
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def app_settings(self) -> dict:
        url = f"{self.base_url}/app_settings"
        response = requests.get(url)
        response.raise_for_status()
        return response.json()

    def documents_info(
        self,
        document_ids: Optional[list[str]] = None,
        user_ids: Optional[list[str]] = None,
    ) -> dict:
        url = f"{self.base_url}/documents_info"
        params = {}
        if document_ids is not None:
            params["document_ids"] = ",".join(document_ids)
        if user_ids is not None:
            params["user_ids"] = ",".join(user_ids)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def analytics(self, filter_criteria: dict, analysis_types: dict) -> dict:
        url = f"{self.base_url}/analytics"
        data = {
            "filter_criteria": filter_criteria,
            "analysis_types": analysis_types,
        }

        try:
            response = requests.post(url, json=data)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            if e.response is None:
                raise requests.exceptions.RequestException(
                    f"Error occurred while calling analytics API. {str(e)}"
                ) from e
            status_code = e.response.status_code
            error_message = e.response.text
            raise requests.exceptions.RequestException(
                f"Error occurred while calling analytics API. Status Code: {status_code}, Error Message: {error_message}"
            ) from e

    def users_stats(self, user_ids: Optional[list[str]] = None) -> dict:
        url = f"{self.base_url}/users_stats"
        params = {}
        if user_ids is not None:
            params["user_ids"] = ",".join(user_ids)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def documents_info(
        self,
        document_ids: Optional[str] = None,
        user_ids: Optional[str] = None,
    ) -> dict:
        url = f"{self.base_url}/documents_info"
        params = {}
        params["document_ids"] = (
            json.dumps(document_ids) if document_ids else None
        )
        params["user_ids"] = json.dumps(user_ids) if user_ids else None
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def document_chunks(self, document_id: str) -> dict:
        url = f"{self.base_url}/document_chunks"
        params = {"document_id": document_id}
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def users_stats(self, user_ids: Optional[list[str]] = None) -> dict:
        url = f"{self.base_url}/users_stats"
        params = {}
        if user_ids is not None:
            params["user_ids"] = ",".join(user_ids)
        response = requests.get(url, params=params)
        response.raise_for_status()
        return response.json()



"""
The `vecs.experimental.adapter.noop` module provides a default no-op (no operation) adapter
that passes the inputs through without any modification. This can be useful when no specific
adapter processing is required.

All public classes, enums, and functions are re-exported by `vecs.adapters` module.
"""

from typing import Any, Dict, Generator, Iterable, Optional, Tuple

from .base import AdapterContext, AdapterStep


class NoOp(AdapterStep):
    """
    NoOp is a no-operation AdapterStep. It is a default adapter that passes through
    the input records without any modifications.
    """

    def __init__(self, dimension: int):
        """
        Initializes the NoOp adapter with a dimension.

        Args:
            dimension (int): The dimension of the input vectors.
        """
        self._dimension = dimension

    @property
    def exported_dimension(self) -> Optional[int]:
        """
        Returns the dimension of the adapter.

        Returns:
            int: The dimension of the input vectors.
        """
        return self._dimension

    def __call__(
        self,
        records: Iterable[Tuple[str, Any, Optional[Dict]]],
        adapter_context: AdapterContext,
    ) -> Generator[Tuple[str, Any, Dict], None, None]:
        """
        Yields the input records without any modification.

        Args:
            records: Iterable of tuples each containing an id, a media and an optional dict.
            adapter_context: Context of the adapter.

        Yields:
            Tuple[str, Any, Dict]: The input record.
        """
        for id, media, metadata in records:
            yield (id, media, metadata or {})



from .base_parser import AsyncParser
from .parser_impls import (
    AudioParser,
    CSVParser,
    DOCXParser,
    HTMLParser,
    ImageParser,
    JSONParser,
    MarkdownParser,
    MovieParser,
    PDFParser,
    PPTParser,
    TextParser,
    XLSXParser,
)

__all__ = [
    "AsyncParser",
    "AudioParser",
    "CSVParser",
    "DOCXParser",
    "HTMLParser",
    "ImageParser",
    "JSONParser",
    "MarkdownParser",
    "MovieParser",
    "PDFParser",
    "PPTParser",
    "ReductoParser",
    "TextParser",
    "XLSXParser",
]



import asyncio
import copy
import json
import logging
import uuid
from abc import abstractmethod
from typing import Any, AsyncGenerator, Optional

from aiohttp import ClientError

from r2r.core import (
    AsyncState,
    Extraction,
    Fragment,
    FragmentType,
    GenerationConfig,
    KGExtraction,
    KGProvider,
    KVLoggingSingleton,
    LLMProvider,
    LoggableAsyncPipe,
    PipeType,
    PromptProvider,
    TextSplitter,
    VectorEntry,
    extract_entities,
    extract_triples,
    generate_id_from_label,
)

logger = logging.getLogger(__name__)


class KGPipe(LoggableAsyncPipe):
    class Input(LoggableAsyncPipe.Input):
        message: AsyncGenerator[Extraction, None]

    def __init__(
        self,
        kg_provider: KGProvider,
        prompt_provider: PromptProvider,
        llm_provider: LLMProvider,
        pipe_logger: Optional[KVLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        config: Optional[LoggableAsyncPipe.PipeConfig] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            pipe_logger=pipe_logger,
            type=type,
            config=config,
            *args,
            **kwargs,
        )
        self.kg_provider = kg_provider
        self.prompt_provider = prompt_provider
        self.llm_provider = llm_provider

    @abstractmethod
    async def fragment(
        self, extraction: Extraction
    ) -> AsyncGenerator[Fragment, None]:
        pass

    @abstractmethod
    async def transform_fragments(
        self, fragments: list[Fragment], metadatas: list[dict]
    ) -> AsyncGenerator[Fragment, None]:
        pass

    @abstractmethod
    async def extract_kg(
        self,
        fragments: list[Fragment],
        kg_generation_config: GenerationConfig,
        *args,
        **kwargs,
    ) -> AsyncGenerator[Fragment, None]:
        pass

    @abstractmethod
    async def _run_logic(
        self,
        input: AsyncGenerator[Extraction, None],
        state: AsyncState,
        run_id: uuid.UUID,
        kg_generation_config: GenerationConfig = GenerationConfig(
            model="gpt-4o"
        ),
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[VectorEntry, None]:
        pass


class R2RKGPipe(KGPipe):
    """
    Embeds and stores documents using a specified embedding model and database.
    """

    def __init__(
        self,
        kg_provider: KGProvider,
        llm_provider: LLMProvider,
        prompt_provider: PromptProvider,
        text_splitter: TextSplitter,
        kg_batch_size: int = 1,
        id_prefix: str = "demo",
        pipe_logger: Optional[KVLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        config: Optional[LoggableAsyncPipe.PipeConfig] = None,
        *args,
        **kwargs,
    ):
        """
        Initializes the embedding pipe with necessary components and configurations.
        """
        super().__init__(
            kg_provider=kg_provider,
            prompt_provider=prompt_provider,
            llm_provider=llm_provider,
            pipe_logger=pipe_logger,
            type=type,
            config=config
            or LoggableAsyncPipe.PipeConfig(name="default_embedding_pipe"),
        )
        self.text_splitter = text_splitter
        self.kg_batch_size = kg_batch_size
        self.id_prefix = id_prefix
        self.pipe_run_info = None

    async def fragment(
        self, extraction: Extraction, run_id: uuid.UUID
    ) -> AsyncGenerator[Fragment, None]:
        """
        Splits text into manageable chunks for embedding.
        """
        if not isinstance(extraction, Extraction):
            raise ValueError(
                f"Expected an Extraction, but received {type(extraction)}."
            )
        if not isinstance(extraction.data, str):
            raise ValueError(
                f"Expected a string, but received {type(extraction.data)}."
            )
        text_chunks = [
            ele.page_content
            for ele in self.text_splitter.create_documents([extraction.data])
        ]
        for iteration, chunk in enumerate(text_chunks):
            fragment = Fragment(
                id=generate_id_from_label(f"{extraction.id}-{iteration}"),
                type=FragmentType.TEXT,
                data=chunk,
                metadata=copy.deepcopy(extraction.metadata),
                extraction_id=extraction.id,
                document_id=extraction.document_id,
            )
            yield fragment

    async def transform_fragments(
        self, fragments: list[Fragment], metadatas: list[dict]
    ) -> AsyncGenerator[Fragment, None]:
        """
        Transforms text chunks based on their metadata, e.g., adding prefixes.
        """
        async for fragment, metadata in zip(fragments, metadatas):
            if "chunk_prefix" in metadata:
                prefix = metadata.pop("chunk_prefix")
                fragment.data = f"{prefix}\n{fragment.data}"
            yield fragment

    async def extract_kg(
        self,
        fragment: Fragment,
        kg_generation_config: GenerationConfig,
        retries: int = 3,
        delay: int = 2,
    ) -> KGExtraction:
        """
        Extracts NER triples from a list of fragments with retries.
        """
        task_prompt = self.prompt_provider.get_prompt(
            self.kg_provider.config.kg_extraction_prompt,
            inputs={"input": fragment.data},
        )
        messages = self.prompt_provider._get_message_payload(
            self.prompt_provider.get_prompt("default_system"), task_prompt
        )
        for attempt in range(retries):
            try:
                response = await self.llm_provider.aget_completion(
                    messages, kg_generation_config
                )

                kg_extraction = response.choices[0].message.content

                # Parsing JSON from the response
                kg_json = json.loads(
                    kg_extraction.split("```json")[1].split("```")[0]
                )

                llm_payload = kg_json.get("entities_and_triplets", {})

                # Extract triples with detailed logging
                entities = extract_entities(llm_payload)

                triples = extract_triples(llm_payload, entities)
                # Create KG extraction object
                return KGExtraction(entities=entities, triples=triples)
            except (
                ClientError,
                json.JSONDecodeError,
                KeyError,
                IndexError,
            ) as e:
                logger.error(f"Error in extract_kg: {e}")
                if attempt < retries - 1:
                    await asyncio.sleep(delay)
                else:
                    logger.error(f"Failed after retries with {e}")
                    # raise e  # Ensure the exception is raised after the final attempt

        return KGExtraction(entities=[], triples=[])

    async def _process_batch(
        self,
        fragment_batch: list[Fragment],
        kg_generation_config: GenerationConfig,
    ) -> list[KGExtraction]:
        """
        Embeds a batch of fragments and yields vector entries.
        """
        tasks = [
            asyncio.create_task(
                self.extract_kg(fragment, kg_generation_config)
            )
            for fragment in fragment_batch
        ]
        return await asyncio.gather(*tasks)

    async def _run_logic(
        self,
        input: KGPipe.Input,
        state: AsyncState,
        run_id: uuid.UUID,
        kg_generation_config: GenerationConfig = GenerationConfig(
            model="gpt-4o", temperature=0.0
        ),
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[KGExtraction, None]:
        """
        Executes the embedding pipe: chunking, transforming, embedding, and storing documents.
        """
        batch_tasks = []
        fragment_batch = []

        fragment_info = {}
        async for extraction in input.message:
            async for fragment in self.fragment(extraction, run_id):
                if extraction.document_id in fragment_info:
                    fragment_info[extraction.document_id] += 1
                else:
                    fragment_info[extraction.document_id] = 1
                extraction.metadata["chunk_order"] = fragment_info[
                    extraction.document_id
                ]
                fragment_batch.append(fragment)
                if len(fragment_batch) >= self.kg_batch_size:
                    # Here, ensure `_process_batch` is scheduled as a coroutine, not called directly
                    batch_tasks.append(
                        self._process_batch(
                            fragment_batch.copy(), kg_generation_config
                        )
                    )  # pass a copy if necessary
                    fragment_batch.clear()  # Clear the batch for new fragments

        logger.info(
            f"Fragmented the input document ids into counts as shown: {fragment_info}"
        )

        if fragment_batch:  # Process any remaining fragments
            batch_tasks.append(self._process_batch(fragment_batch.copy()))

        # Process tasks as they complete
        for task in asyncio.as_completed(batch_tasks):
            batch_result = await task  # Wait for the next task to complete
            for kg_extraction in batch_result:
                yield kg_extraction



# abstractions are taken from LlamaIndex
# https://github.com/run-llama/llama_index
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, StrictFloat, StrictInt, StrictStr


class LabelledNode(BaseModel):
    """An entity in a graph."""

    label: str = Field(default="node", description="The label of the node.")
    embedding: Optional[List[float]] = Field(
        default=None, description="The embeddings of the node."
    )
    properties: Dict[str, Any] = Field(default_factory=dict)

    @abstractmethod
    def __str__(self) -> str:
        """Return the string representation of the node."""
        ...

    @property
    @abstractmethod
    def id(self) -> str:
        """Get the node id."""
        ...


class EntityNode(LabelledNode):
    """An entity in a graph."""

    name: str = Field(description="The name of the entity.")
    label: str = Field(default="entity", description="The label of the node.")
    properties: Dict[str, Any] = Field(default_factory=dict)

    def __str__(self) -> str:
        """Return the string representation of the node."""
        return self.name

    @property
    def id(self) -> str:
        """Get the node id."""
        return self.name.replace('"', " ")


class ChunkNode(LabelledNode):
    """A text chunk in a graph."""

    text: str = Field(description="The text content of the chunk.")
    id_: Optional[str] = Field(
        default=None,
        description="The id of the node. Defaults to a hash of the text.",
    )
    label: str = Field(
        default="text_chunk", description="The label of the node."
    )
    properties: Dict[str, Any] = Field(default_factory=dict)

    def __str__(self) -> str:
        """Return the string representation of the node."""
        return self.text

    @property
    def id(self) -> str:
        """Get the node id."""
        return str(hash(self.text)) if self.id_ is None else self.id_


class Relation(BaseModel):
    """A relation connecting two entities in a graph."""

    label: str
    source_id: str
    target_id: str
    properties: Dict[str, Any] = Field(default_factory=dict)

    def __str__(self) -> str:
        """Return the string representation of the relation."""
        return self.label

    @property
    def id(self) -> str:
        """Get the relation id."""
        return self.label


Triplet = Tuple[LabelledNode, Relation, LabelledNode]


class VectorStoreQueryMode(str, Enum):
    """Vector store query mode."""

    DEFAULT = "default"
    SPARSE = "sparse"
    HYBRID = "hybrid"
    TEXT_SEARCH = "text_search"
    SEMANTIC_HYBRID = "semantic_hybrid"

    # fit learners
    SVM = "svm"
    LOGISTIC_REGRESSION = "logistic_regression"
    LINEAR_REGRESSION = "linear_regression"

    # maximum marginal relevance
    MMR = "mmr"


class FilterOperator(str, Enum):
    """Vector store filter operator."""

    # TODO add more operators
    EQ = "=="  # default operator (string, int, float)
    GT = ">"  # greater than (int, float)
    LT = "<"  # less than (int, float)
    NE = "!="  # not equal to (string, int, float)
    GTE = ">="  # greater than or equal to (int, float)
    LTE = "<="  # less than or equal to (int, float)
    IN = "in"  # In array (string or number)
    NIN = "nin"  # Not in array (string or number)
    ANY = "any"  # Contains any (array of strings)
    ALL = "all"  # Contains all (array of strings)
    TEXT_MATCH = "text_match"  # full text match (allows you to search for a specific substring, token or phrase within the text field)
    CONTAINS = "contains"  # metadata array contains value (string or number)


class MetadataFilter(BaseModel):
    """Comprehensive metadata filter for vector stores to support more operators.

    Value uses Strict* types, as int, float and str are compatible types and were all
    converted to string before.

    See: https://docs.pydantic.dev/latest/usage/types/#strict-types
    """

    key: str
    value: Union[
        StrictInt,
        StrictFloat,
        StrictStr,
        List[StrictStr],
        List[StrictFloat],
        List[StrictInt],
    ]
    operator: FilterOperator = FilterOperator.EQ

    @classmethod
    def from_dict(
        cls,
        filter_dict: Dict,
    ) -> "MetadataFilter":
        """Create MetadataFilter from dictionary.

        Args:
            filter_dict: Dict with key, value and operator.

        """
        return MetadataFilter.parse_obj(filter_dict)


# # TODO: Deprecate ExactMatchFilter and use MetadataFilter instead
# # Keep class for now so that AutoRetriever can still work with old vector stores
# class ExactMatchFilter(BaseModel):
#     key: str
#     value: Union[StrictInt, StrictFloat, StrictStr]

# set ExactMatchFilter to MetadataFilter
ExactMatchFilter = MetadataFilter


class FilterCondition(str, Enum):
    """Vector store filter conditions to combine different filters."""

    # TODO add more conditions
    AND = "and"
    OR = "or"


class MetadataFilters(BaseModel):
    """Metadata filters for vector stores."""

    # Exact match filters and Advanced filters with operators like >, <, >=, <=, !=, etc.
    filters: List[Union[MetadataFilter, ExactMatchFilter, "MetadataFilters"]]
    # and/or such conditions for combining different filters
    condition: Optional[FilterCondition] = FilterCondition.AND


@dataclass
class VectorStoreQuery:
    """Vector store query."""

    query_embedding: Optional[List[float]] = None
    similarity_top_k: int = 1
    doc_ids: Optional[List[str]] = None
    node_ids: Optional[List[str]] = None
    query_str: Optional[str] = None
    output_fields: Optional[List[str]] = None
    embedding_field: Optional[str] = None

    mode: VectorStoreQueryMode = VectorStoreQueryMode.DEFAULT

    # NOTE: only for hybrid search (0 for bm25, 1 for vector search)
    alpha: Optional[float] = None

    # metadata filters
    filters: Optional[MetadataFilters] = None

    # only for mmr
    mmr_threshold: Optional[float] = None

    # NOTE: currently only used by postgres hybrid search
    sparse_top_k: Optional[int] = None
    # NOTE: return top k results from hybrid search. similarity_top_k is used for dense search top k
    hybrid_top_k: Optional[int] = None


class PropertyGraphStore(ABC):
    """Abstract labelled graph store protocol.

    This protocol defines the interface for a graph store, which is responsible
    for storing and retrieving knowledge graph data.

    Attributes:
        client: Any: The client used to connect to the graph store.
        get: Callable[[str], List[List[str]]]: Get triplets for a given subject.
        get_rel_map: Callable[[Optional[List[str]], int], Dict[str, List[List[str]]]]:
            Get subjects' rel map in max depth.
        upsert_triplet: Callable[[str, str, str], None]: Upsert a triplet.
        delete: Callable[[str, str, str], None]: Delete a triplet.
        persist: Callable[[str, Optional[fsspec.AbstractFileSystem]], None]:
            Persist the graph store to a file.
    """

    supports_structured_queries: bool = False
    supports_vector_queries: bool = False

    @property
    def client(self) -> Any:
        """Get client."""
        ...

    @abstractmethod
    def get(
        self,
        properties: Optional[dict] = None,
        ids: Optional[List[str]] = None,
    ) -> List[LabelledNode]:
        """Get nodes with matching values."""
        ...

    @abstractmethod
    def get_triplets(
        self,
        entity_names: Optional[List[str]] = None,
        relation_names: Optional[List[str]] = None,
        properties: Optional[dict] = None,
        ids: Optional[List[str]] = None,
    ) -> List[Triplet]:
        """Get triplets with matching values."""
        ...

    @abstractmethod
    def get_rel_map(
        self,
        graph_nodes: List[LabelledNode],
        depth: int = 2,
        limit: int = 30,
        ignore_rels: Optional[List[str]] = None,
    ) -> List[Triplet]:
        """Get depth-aware rel map."""
        ...

    @abstractmethod
    def upsert_nodes(self, nodes: List[LabelledNode]) -> None:
        """Upsert nodes."""
        ...

    @abstractmethod
    def upsert_relations(self, relations: List[Relation]) -> None:
        """Upsert relations."""
        ...

    @abstractmethod
    def delete(
        self,
        entity_names: Optional[List[str]] = None,
        relation_names: Optional[List[str]] = None,
        properties: Optional[dict] = None,
        ids: Optional[List[str]] = None,
    ) -> None:
        """Delete matching data."""
        ...

    @abstractmethod
    def structured_query(
        self, query: str, param_map: Optional[Dict[str, Any]] = None
    ) -> Any:
        """Query the graph store with statement and parameters."""
        ...

    @abstractmethod
    def vector_query(
        self, query: VectorStoreQuery, **kwargs: Any
    ) -> Tuple[List[LabelledNode], List[float]]:
        """Query the graph store with a vector store query."""
        ...

    # def persist(
    #     self, persist_path: str, fs: Optional[fsspec.AbstractFileSystem] = None
    # ) -> None:
    #     """Persist the graph store to a file."""
    #     return

    def get_schema(self, refresh: bool = False) -> Any:
        """Get the schema of the graph store."""
        return None

    def get_schema_str(self, refresh: bool = False) -> str:
        """Get the schema of the graph store as a string."""
        return str(self.get_schema(refresh=refresh))

    ### ----- Async Methods ----- ###

    async def aget(
        self,
        properties: Optional[dict] = None,
        ids: Optional[List[str]] = None,
    ) -> List[LabelledNode]:
        """Asynchronously get nodes with matching values."""
        return self.get(properties, ids)

    async def aget_triplets(
        self,
        entity_names: Optional[List[str]] = None,
        relation_names: Optional[List[str]] = None,
        properties: Optional[dict] = None,
        ids: Optional[List[str]] = None,
    ) -> List[Triplet]:
        """Asynchronously get triplets with matching values."""
        return self.get_triplets(entity_names, relation_names, properties, ids)

    async def aget_rel_map(
        self,
        graph_nodes: List[LabelledNode],
        depth: int = 2,
        limit: int = 30,
        ignore_rels: Optional[List[str]] = None,
    ) -> List[Triplet]:
        """Asynchronously get depth-aware rel map."""
        return self.get_rel_map(graph_nodes, depth, limit, ignore_rels)

    async def aupsert_nodes(self, nodes: List[LabelledNode]) -> None:
        """Asynchronously add nodes."""
        return self.upsert_nodes(nodes)

    async def aupsert_relations(self, relations: List[Relation]) -> None:
        """Asynchronously add relations."""
        return self.upsert_relations(relations)

    async def adelete(
        self,
        entity_names: Optional[List[str]] = None,
        relation_names: Optional[List[str]] = None,
        properties: Optional[dict] = None,
        ids: Optional[List[str]] = None,
    ) -> None:
        """Asynchronously delete matching data."""
        return self.delete(entity_names, relation_names, properties, ids)

    async def astructured_query(
        self, query: str, param_map: Optional[Dict[str, Any]] = {}
    ) -> Any:
        """Asynchronously query the graph store with statement and parameters."""
        return self.structured_query(query, param_map)

    async def avector_query(
        self, query: VectorStoreQuery, **kwargs: Any
    ) -> Tuple[List[LabelledNode], List[float]]:
        """Asynchronously query the graph store with a vector store query."""
        return self.vector_query(query, **kwargs)

    async def aget_schema(self, refresh: bool = False) -> str:
        """Asynchronously get the schema of the graph store."""
        return self.get_schema(refresh=refresh)

    async def aget_schema_str(self, refresh: bool = False) -> str:
        """Asynchronously get the schema of the graph store as a string."""
        return str(await self.aget_schema(refresh=refresh))


LIST_LIMIT = 128


def clean_string_values(text: str) -> str:
    return text.replace("\n", " ").replace("\r", " ")


def value_sanitize(d: Any) -> Any:
    """Sanitize the input dictionary or list.

    Sanitizes the input by removing embedding-like values,
    lists with more than 128 elements, that are mostly irrelevant for
    generating answers in a LLM context. These properties, if left in
    results, can occupy significant context space and detract from
    the LLM's performance by introducing unnecessary noise and cost.
    """
    if isinstance(d, dict):
        new_dict = {}
        for key, value in d.items():
            if isinstance(value, dict):
                sanitized_value = value_sanitize(value)
                if (
                    sanitized_value is not None
                ):  # Check if the sanitized value is not None
                    new_dict[key] = sanitized_value
            elif isinstance(value, list):
                if len(value) < LIST_LIMIT:
                    sanitized_value = value_sanitize(value)
                    if (
                        sanitized_value is not None
                    ):  # Check if the sanitized value is not None
                        new_dict[key] = sanitized_value
                # Do not include the key if the list is oversized
            else:
                new_dict[key] = value
        return new_dict
    elif isinstance(d, list):
        if len(d) < LIST_LIMIT:
            return [
                value_sanitize(item)
                for item in d
                if value_sanitize(item) is not None
            ]
        else:
            return None
    else:
        return d



# import json
# import logging
# import os
# import uuid
# from typing import Optional, Union

# from sqlalchemy import text

# from r2r.core import (
#     DocumentInfo,
#     SearchResult,
#     UserStats,
#     VectorDBConfig,
#     VectorDBProvider,
#     VectorEntry,
# )
# from r2r.vecs.client import Client
# from r2r.vecs.collection import Collection

# logger = logging.getLogger(__name__)


# class PGVectorDB(VectorDBProvider):
#     def __init__(self, config: VectorDBConfig) -> None:
#         super().__init__(config)
#         try:
#             import r2r.vecs
#         except ImportError:
#             raise ValueError(
#                 f"Error, PGVectorDB requires the vecs library. Please run `poetry add vecs`."
#             )
#         user = os.getenv("POSTGRES_USER")
#         password = os.getenv("POSTGRES_PASSWORD")
#         host = os.getenv("POSTGRES_HOST")
#         port = os.getenv("POSTGRES_PORT")
#         db_name = os.getenv("POSTGRES_DBNAME")
#         if not all([user, password, host, port, db_name]):
#             raise ValueError(
#                 "Error, please set the POSTGRES_USER, POSTGRES_PASSWORD, POSTGRES_HOST, POSTGRES_PORT, and POSTGRES_DBNAME environment variables."
#             )
#         try:
#             DB_CONNECTION = (
#                 f"postgresql://{user}:{password}@{host}:{port}/{db_name}"
#             )
#             self.vx: Client = r2r.vecs.create_client(DB_CONNECTION)
#         except Exception as e:
#             raise ValueError(
#                 f"Error {e} occurred while attempting to connect to the pgvector provider with {DB_CONNECTION}."
#             )
#         self.collection: Optional[Collection] = None
#         self.config: VectorDBConfig = config

#     def initialize_collection(self, dimension: int) -> None:
#         self.collection = self.vx.get_or_create_collection(
#             name=self.config.collection_name, dimension=dimension
#         )
#         self._create_document_info_table()
#         self._create_hybrid_search_function()

#     def _create_document_info_table(self):
#         with self.vx.Session() as sess:
#             with sess.begin():
#                 query = f"""
#                 CREATE TABLE IF NOT EXISTS document_info_{self.config.collection_name} (
#                     document_id UUID PRIMARY KEY,
#                     title TEXT,
#                     user_id UUID,
#                     version TEXT,
#                     size_in_bytes INT,
#                     created_at TIMESTAMPTZ DEFAULT NOW(),
#                     updated_at TIMESTAMPTZ DEFAULT NOW(),
#                     metadata JSONB
#                 );
#                 """
#                 sess.execute(text(query))
#                 sess.commit()

#     # def _create_hybrid_search_function(self):
#     #     hybrid_search_function = f"""
#     #     CREATE OR REPLACE FUNCTION hybrid_search_{self.config.collection_name}(
#     #         query_text TEXT,
#     #         query_embedding VECTOR(512),
#     #         match_limit INT,
#     #         full_text_weight FLOAT = 1,
#     #         semantic_weight FLOAT = 1,
#     #         rrf_k INT = 50,
#     #         filter_condition JSONB = NULL
#     #     )
#     #     RETURNS SETOF vecs."{self.config.collection_name}"
#     #     LANGUAGE sql
#     #     AS $$
#     #     WITH full_text AS (
#     #         SELECT
#     #             id,
#     #             ROW_NUMBER() OVER (ORDER BY ts_rank(to_tsvector('english', metadata->>'text'), websearch_to_tsquery(query_text)) DESC) AS rank_ix
#     #         FROM vecs."{self.config.collection_name}"
#     #         WHERE to_tsvector('english', metadata->>'text') @@ websearch_to_tsquery(query_text)
#     #         AND (filter_condition IS NULL OR (metadata @> filter_condition))
#     #         ORDER BY rank_ix
#     #         LIMIT LEAST(match_limit, 30) * 2
#     #     ),
#     #     semantic AS (
#     #         SELECT
#     #             id,
#     #             ROW_NUMBER() OVER (ORDER BY vec <#> query_embedding) AS rank_ix
#     #         FROM vecs."{self.config.collection_name}"
#     #         WHERE filter_condition IS NULL OR (metadata @> filter_condition)
#     #         ORDER BY rank_ix
#     #         LIMIT LEAST(match_limit, 30) * 2
#     #     )
#     #     SELECT
#     #         vecs."{self.config.collection_name}".*
#     #     FROM
#     #         full_text
#     #         FULL OUTER JOIN semantic
#     #             ON full_text.id = semantic.id
#     #         JOIN vecs."{self.config.collection_name}"
#     #             ON vecs."{self.config.collection_name}".id = COALESCE(full_text.id, semantic.id)
#     #     ORDER BY
#     #         COALESCE(1.0 / (rrf_k + full_text.rank_ix), 0.0) * full_text_weight +
#     #         COALESCE(1.0 / (rrf_k + semantic.rank_ix), 0.0) * semantic_weight
#     #         DESC
#     #     LIMIT
#     #         LEAST(match_limit, 30);
#     #     $$;
#     #     """
#     #     with self.vx.Session() as sess:
#     #         sess.execute(text(hybrid_search_function))
#     #         sess.commit()


#     def _create_hybrid_search_function(self):
#         table_name = self.config.collection_name
#         hybrid_search_function = f"""
#         CREATE OR REPLACE FUNCTION hybrid_search_{self.config.collection_name}(
#             query_text TEXT,
#             query_embedding VECTOR(512),
#             match_count INT,
#             full_text_weight FLOAT = 1,
#             semantic_weight FLOAT = 1,
#             rrf_k INT = 50
#         )
#         RETURNS SETOF vecs."{self.config.collection_name}"
#         LANGUAGE sql
#         AS $$
#         WITH full_text AS (
#             SELECT
#                 id,
#                 ROW_NUMBER() OVER (ORDER BY ts_rank(to_tsvector('english', metadata->>'text'), websearch_to_tsquery(query_text)) DESC) AS rank_ix
#             FROM vecs."{self.config.collection_name}"
#             WHERE to_tsvector('english', metadata->>'text') @@ websearch_to_tsquery(query_text)
#             ORDER BY rank_ix
#             LIMIT LEAST(match_count, 30) * 2
#         ),
#         semantic AS (
#             SELECT
#                 id,
#                 ROW_NUMBER() OVER (ORDER BY vec <#> query_embedding) AS rank_ix
#             FROM vecs."{self.config.collection_name}"
#             ORDER BY rank_ix
#             LIMIT LEAST(match_count, 30) * 2
#         )
#         SELECT
#             vecs."{self.config.collection_name}".*
#         FROM
#             full_text
#             FULL OUTER JOIN semantic
#                 ON full_text.id = semantic.id
#             JOIN vecs."{self.config.collection_name}"
#                 ON vecs."{self.config.collection_name}".id = COALESCE(full_text.id, semantic.id)
#         ORDER BY
#             COALESCE(1.0 / (rrf_k + full_text.rank_ix), 0.0) * full_text_weight +
#             COALESCE(1.0 / (rrf_k + semantic.rank_ix), 0.0) * semantic_weight
#             DESC
#         LIMIT
#             LEAST(match_count, 30);
#         $$;
#         """
#         with self.vx.Session() as sess:
#             sess.execute(text(hybrid_search_function))
#             sess.commit()

#     def copy(self, entry: VectorEntry, commit=True) -> None:
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `copy`."
#             )

#         serializeable_entry = entry.to_serializable()

#         self.collection.copy(
#             records=[
#                 (
#                     serializeable_entry["id"],
#                     serializeable_entry["vector"],
#                     serializeable_entry["metadata"],
#                 )
#             ]
#         )

#     def copy_entries(
#         self, entries: list[VectorEntry], commit: bool = True
#     ) -> None:
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `copy_entries`."
#             )

#         self.collection.copy(
#             records=[
#                 (
#                     str(entry.id),
#                     entry.vector.data,
#                     entry.to_serializable()["metadata"],
#                 )
#                 for entry in entries
#             ]
#         )

#     def upsert(self, entry: VectorEntry, commit=True) -> None:
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `upsert`."
#             )

#         self.collection.upsert(
#             records=[
#                 (
#                     str(entry.id),
#                     entry.vector.data,
#                     entry.to_serializable()["metadata"],
#                 )
#             ]
#         )

#     def upsert_entries(
#         self, entries: list[VectorEntry], commit: bool = True
#     ) -> None:
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `upsert_entries`."
#             )

#         self.collection.upsert(
#             records=[
#                 (
#                     str(entry.id),
#                     entry.vector.data,
#                     entry.to_serializable()["metadata"],
#                 )
#                 for entry in entries
#             ]
#         )

#     def search(
#         self,
#         query_vector: list[float],
#         filters: dict[str, Union[bool, int, str]] = {},
#         limit: int = 10,
#         *args,
#         **kwargs,
#     ) -> list[SearchResult]:
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `search`."
#             )
#         measure = kwargs.get("measure", "cosine_distance")
#         mapped_filters = {
#             key: {"$eq": value} for key, value in filters.items()
#         }

#         return [
#             SearchResult(id=ele[0], score=float(1 - ele[1]), metadata=ele[2])  # type: ignore
#             for ele in self.collection.query(
#                 data=query_vector,
#                 limit=limit,
#                 filters=mapped_filters,
#                 measure=measure,
#                 include_value=True,
#                 include_metadata=True,
#             )
#         ]

#     def hybrid_search(
#         self,
#         query_text: str,
#         query_vector: list[float],
#         limit: int = 10,
#         # filters: Optional[dict[str, Union[bool, int, str]]] = None,
#         # Hybrid search parameters
#         full_text_weight: float = 1.0,
#         semantic_weight: float = 1.0,
#         rrf_k: int = 20,  # typical value is ~2x the number of results you want
#         *args,
#         **kwargs,
#     ) -> list[SearchResult]:
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `hybrid_search`."
#             )

#         # Convert filters to a JSON-compatible format
#         filter_condition = None
#         # if filters:
#         #     filter_condition = json.dumps(filters)
#         query = text(
#             f"""
#             SELECT * FROM hybrid_search_{self.config.collection_name}(
#                 cast(:query_text as TEXT), cast(:query_embedding as VECTOR), cast(:match_limit as INT),
#                 cast(:full_text_weight as FLOAT), cast(:semantic_weight as FLOAT), cast(:rrf_k as INT)
#             )
#         """
#         )

#         # query = text(
#         #     f"""
#         #     SELECT * FROM hybrid_search_{self.config.collection_name}(
#         #         cast(:query_text as TEXT), cast(:query_embedding as VECTOR), cast(:match_limit as INT),
#         #         cast(:full_text_weight as FLOAT), cast(:semantic_weight as FLOAT), cast(:rrf_k as INT),
#         #         cast(:filter_condition as JSONB)
#         #     )
#         # """
#         # )

#         params = {
#             "query_text": str(query_text),
#             "query_embedding": list(query_vector),
#             "match_limit": limit,
#             "full_text_weight": full_text_weight,
#             "semantic_weight": semantic_weight,
#             "rrf_k": rrf_k,
#             "filter_condition": filter_condition,
#         }

#         with self.vx.Session() as session:
#             print('doing hybrid search with params = ', params)
#             result = session.execute(query, params).fetchall()
#         return [
#             SearchResult(id=row[0], score=1.0, metadata=row[-1])
#             for row in result
#         ]

#     def create_index(self, index_type, column_name, index_options):
#         pass

#     def delete_by_metadata(
#         self, metadata_fields: str, metadata_values: Union[bool, int, str]
#     ) -> list[str]:
#         super().delete_by_metadata(metadata_fields, metadata_values)
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `delete_by_metadata`."
#             )
#         return self.collection.delete(
#             filters={
#                 k: {"$eq": v} for k, v in zip(metadata_fields, metadata_values)
#             }
#         )

#     def get_metadatas(
#         self,
#         metadata_fields: list[str],
#         filter_field: Optional[str] = None,
#         filter_value: Optional[Union[bool, int, str]] = None,
#     ) -> list[dict]:
#         if self.collection is None:
#             raise ValueError(
#                 "Please call `initialize_collection` before attempting to run `get_metadatas`."
#             )

#         results = {tuple(metadata_fields): {}}
#         for field in metadata_fields:
#             unique_values = self.collection.get_unique_metadata_values(
#                 field=field,
#                 filter_field=filter_field,
#                 filter_value=filter_value,
#             )
#             for value in unique_values:
#                 if value not in results:
#                     results[value] = {}
#                 results[value][field] = value

#         return [
#             results[key] for key in results if key != tuple(metadata_fields)
#         ]

#     def upsert_documents_info(
#         self, documents_info: list[DocumentInfo]
#     ) -> None:
#         for document_info in documents_info:
#             db_entry = document_info.convert_to_db_entry()
#             query = text(
#                 f"""
#                 INSERT INTO document_info_{self.config.collection_name} (document_id, title, user_id, version, created_at, updated_at, size_in_bytes, metadata)
#                 VALUES (:document_id, :title, :user_id, :version, :created_at, :updated_at, :size_in_bytes, :metadata)
#                 ON CONFLICT (document_id) DO UPDATE SET
#                     title = EXCLUDED.title,
#                     user_id = EXCLUDED.user_id,
#                     version = EXCLUDED.version,
#                     updated_at = EXCLUDED.updated_at,
#                     size_in_bytes = EXCLUDED.size_in_bytes,
#                     metadata = EXCLUDED.metadata;
#             """
#             )
#         with self.vx.Session() as sess:
#             sess.execute(query, db_entry)
#             sess.commit()

#     def delete_documents_info(self, document_ids: list[str]) -> None:
#         placeholders = ", ".join(
#             f":doc_id_{i}" for i in range(len(document_ids))
#         )
#         query = text(
#             f"""
#             DELETE FROM document_info_{self.config.collection_name} WHERE document_id IN ({placeholders});
#             """
#         )
#         params = {
#             f"doc_id_{i}": document_id
#             for i, document_id in enumerate(document_ids)
#         }

#         with self.vx.Session() as sess:
#             with sess.begin():
#                 sess.execute(query, params)
#             sess.commit()

#     def get_documents_info(
#         self,
#         filter_document_ids: Optional[list[str]] = None,
#         filter_user_ids: Optional[list[str]] = None,
#     ):
#         conditions = []
#         params = {}

#         if filter_document_ids:
#             placeholders = ", ".join(
#                 f":doc_id_{i}" for i in range(len(filter_document_ids))
#             )
#             conditions.append(f"document_id IN ({placeholders})")
#             params.update(
#                 {
#                     f"doc_id_{i}": str(document_id)
#                     for i, document_id in enumerate(filter_document_ids)
#                 }
#             )
#         if filter_user_ids:
#             placeholders = ", ".join(
#                 f":user_id_{i}" for i in range(len(filter_user_ids))
#             )
#             conditions.append(f"user_id IN ({placeholders})")
#             params.update(
#                 {
#                     f"user_id_{i}": str(user_id)
#                     for i, user_id in enumerate(filter_user_ids)
#                 }
#             )

#         query = f"""
#             SELECT document_id, title, user_id, version, size_in_bytes, created_at, updated_at, metadata
#             FROM document_info_{self.config.collection_name}
#         """
#         if conditions:
#             query += " WHERE " + " AND ".join(conditions)

#         with self.vx.Session() as sess:
#             results = sess.execute(text(query), params).fetchall()
#             return [
#                 DocumentInfo(
#                     document_id=row[0],
#                     title=row[1],
#                     user_id=row[2],
#                     version=row[3],
#                     size_in_bytes=row[4],
#                     created_at=row[5],
#                     updated_at=row[6],
#                     metadata=row[7],
#                 )
#                 for row in results
#             ]

#     def get_document_chunks(self, document_id: str) -> list[dict]:
#         if not self.collection:
#             raise ValueError("Collection is not initialized.")

#         table_name = self.collection.table.name
#         query = text(
#             f"""
#             SELECT metadata
#             FROM vecs."{table_name}"
#             WHERE metadata->>'document_id' = :document_id
#             ORDER BY CAST(metadata->>'chunk_order' AS INTEGER)
#         """
#         )

#         params = {"document_id": document_id}

#         with self.vx.Session() as sess:
#             results = sess.execute(query, params).fetchall()
#             return [result[0] for result in results]

#     def get_users_stats(self, user_ids: Optional[list[str]] = None):
#         user_ids_condition = ""
#         params = {}
#         if user_ids:
#             user_ids_condition = "WHERE user_id IN :user_ids"
#             params["user_ids"] = tuple(
#                 map(str, user_ids)
#             )  # Convert UUIDs to strings

#         query = f"""
#             SELECT user_id, COUNT(document_id) AS num_files, SUM(size_in_bytes) AS total_size_in_bytes, ARRAY_AGG(document_id) AS document_ids
#             FROM document_info_{self.config.collection_name}
#             {user_ids_condition}
#             GROUP BY user_id
#         """

#         with self.vx.Session() as sess:
#             results = sess.execute(text(query), params).fetchall()

#         return [
#             UserStats(
#                 user_id=row[0],
#                 num_files=row[1],
#                 total_size_in_bytes=row[2],
#                 document_ids=[uuid.UUID(doc_id) for doc_id in row[3]],
#             )
#             for row in results
#         ]



import os
from typing import Optional


class ExaClient:
    def __init__(self, api_key: Optional[str] = None) -> None:
        try:
            from exa_py import Exa
        except ImportError:
            raise ImportError(
                "Please install the `exa-py` package to use `ExaClient`."
            )

        api_key = api_key or os.getenv("EXA_API_KEY")

        if not api_key:
            raise ValueError(
                "Please set the `EXA_API_KEY` environment variable to use `ExaClient`."
            )

        self.exa = Exa(api_key)

    def _search_and_contents(
        self,
        query: str,
        limit: int = 10,
        num_highlights: int = 2,
        num_sentences: int = 4,
        **kwargs
    ) -> list:
        return self.exa.search_and_contents(
            query,
            num_results=limit,
            highlights={
                "highlights_per_url": num_highlights,
                "num_sentences": num_sentences,
                "query": query,
            },
            use_autoprompt=True,
            **kwargs
        )

    def get_raw(self, query: str, limit: int = 10, **kwargs) -> list:
        search = self._search_and_contents(query, limit, **kwargs)

        # Transform to be consistent w/ Serper API
        xf_results = []
        for i, result in enumerate(search.results):
            xf_results.append(
                {
                    "title": result.title,
                    "link": result.url,
                    "snippet": ".".join(result.highlights),
                    "position": i + 1,
                    "type": "organic",
                }
            )
        return xf_results

    @staticmethod
    def construct_context(results: list) -> str:
        from .serper import SerperClient

        return SerperClient.construct_context(results)



"""Abstractions for documents and their extractions."""

import json
import logging
import uuid
from datetime import datetime
from enum import Enum
from typing import Optional, Union

from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)

DataType = Union[str, bytes]


class DocumentType(Enum):
    """Types of documents that can be stored."""

    CSV = "csv"
    DOCX = "docx"
    HTML = "html"
    JSON = "json"
    MD = "md"
    PDF = "pdf"
    PPTX = "pptx"
    TXT = "txt"
    XLSX = "xlsx"
    GIF = "gif"
    PNG = "png"
    JPG = "jpg"
    JPEG = "jpeg"
    SVG = "svg"
    MP3 = "mp3"
    MP4 = "mp4"


class Document(BaseModel):
    """A document that has been stored in the system."""

    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    type: DocumentType
    data: DataType
    metadata: dict

    title: Optional[str] = None
    user_id: Optional[uuid.UUID] = None


class DocumentInfo(BaseModel):
    """Base class for document information handling."""

    document_id: uuid.UUID
    version: str
    size_in_bytes: int
    metadata: dict

    title: Optional[str] = None
    user_id: Optional[uuid.UUID] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    def convert_to_db_entry(self):
        """Prepare the document info for database entry, extracting certain fields from metadata."""
        now = datetime.now()
        return {
            "document_id": str(self.document_id),
            "title": self.title or "N/A",
            "user_id": str(self.user_id),
            "version": self.version,
            "size_in_bytes": self.size_in_bytes,
            "metadata": json.dumps(self.metadata),
            "created_at": self.created_at or now,
            "updated_at": self.updated_at or now,
        }


class ExtractionType(Enum):
    """Types of extractions that can be performed."""

    TXT = "txt"
    IMG = "img"
    MOV = "mov"


class Extraction(BaseModel):
    """An extraction from a document."""

    id: uuid.UUID
    type: ExtractionType = ExtractionType.TXT
    data: DataType
    metadata: dict
    document_id: uuid.UUID


class FragmentType(Enum):
    """A type of fragment that can be extracted from a document."""

    TEXT = "text"
    IMAGE = "image"


class Fragment(BaseModel):
    """A fragment extracted from a document."""

    id: uuid.UUID
    type: FragmentType
    data: DataType
    metadata: dict
    document_id: uuid.UUID
    extraction_id: uuid.UUID


class Entity(BaseModel):
    """An entity extracted from a document."""

    category: str
    subcategory: Optional[str] = None
    value: str

    def __str__(self):
        return (
            f"{self.category}:{self.subcategory}:{self.value}"
            if self.subcategory
            else f"{self.category}:{self.value}"
        )


class Triple(BaseModel):
    """A triple extracted from a document."""

    subject: str
    predicate: str
    object: str


def extract_entities(llm_payload: list[str]) -> dict[str, Entity]:
    entities = {}
    for entry in llm_payload:
        try:
            if "], " in entry:  # Check if the entry is an entity
                entry_val = entry.split("], ")[0] + "]"
                entry = entry.split("], ")[1]
                colon_count = entry.count(":")

                if colon_count == 1:
                    category, value = entry.split(":")
                    subcategory = None
                elif colon_count >= 2:
                    parts = entry.split(":", 2)
                    category, subcategory, value = (
                        parts[0],
                        parts[1],
                        parts[2],
                    )
                else:
                    raise ValueError("Unexpected entry format")

                entities[entry_val] = Entity(
                    category=category, subcategory=subcategory, value=value
                )
        except Exception as e:
            logger.error(f"Error processing entity {entry}: {e}")
            continue
    return entities


def extract_triples(
    llm_payload: list[str], entities: dict[str, Entity]
) -> list[Triple]:
    triples = []
    for entry in llm_payload:
        try:
            if "], " not in entry:  # Check if the entry is an entity
                subject, predicate, object = entry.split(" ")
                subject = entities[subject].value  # Use entity.value
                if "[" in object and "]" in object:
                    object = entities[object].value  # Use entity.value
                triples.append(
                    Triple(subject=subject, predicate=predicate, object=object)
                )
        except Exception as e:
            logger.error(f"Error processing triplet {entry}: {e}")
            continue
    return triples


class KGExtraction(BaseModel):
    """An extraction from a document that is part of a knowledge graph."""

    entities: dict[str, Entity]
    triples: list[Triple]



import uuid
from abc import abstractmethod
from typing import Any, AsyncGenerator, Optional

from r2r.core import (
    AsyncState,
    GenerationConfig,
    KVLoggingSingleton,
    LLMProvider,
    LoggableAsyncPipe,
    PipeType,
    PromptProvider,
)


class GeneratorPipe(LoggableAsyncPipe):
    class Config(LoggableAsyncPipe.PipeConfig):
        name: str
        task_prompt: str
        system_prompt: str = "default_system"

    def __init__(
        self,
        llm_provider: LLMProvider,
        prompt_provider: PromptProvider,
        type: PipeType = PipeType.GENERATOR,
        config: Optional[Config] = None,
        pipe_logger: Optional[KVLoggingSingleton] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            type=type,
            config=config or self.Config(),
            pipe_logger=pipe_logger,
            *args,
            **kwargs,
        )
        self.llm_provider = llm_provider
        self.prompt_provider = prompt_provider

    @abstractmethod
    async def _run_logic(
        self,
        input: LoggableAsyncPipe.Input,
        state: AsyncState,
        run_id: uuid.UUID,
        rag_generation_config: GenerationConfig,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[Any, None]:
        pass

    @abstractmethod
    def _get_message_payload(
        self, message: str, *args: Any, **kwargs: Any
    ) -> list:
        pass



"""Base classes for knowledge graph providers."""

import logging
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

from ..abstractions.llama_abstractions import (
    EntityNode,
    LabelledNode,
    Relation,
    VectorStoreQuery,
)
from .base_provider import ProviderConfig

logger = logging.getLogger(__name__)


class KGConfig(ProviderConfig):
    """A base KG config class"""

    provider: Optional[str] = None
    batch_size: int = 1
    kg_extraction_prompt: Optional[str] = "ner_kg_extraction"
    kg_agent_prompt: Optional[str] = "kg_agent"

    def validate(self) -> None:
        if not self.provider:
            raise ValueError("Provider must be set.")
        if self.provider and self.provider not in self.supported_providers:
            raise ValueError(f"Provider '{self.provider}' is not supported.")

    @property
    def supported_providers(self) -> list[str]:
        return ["None", "neo4j"]


class KGProvider(ABC):
    """An abstract class to provide a common interface for Knowledge Graphs."""

    def __init__(self, config: KGConfig) -> None:
        if not isinstance(config, KGConfig):
            raise ValueError(
                "KGProvider must be initialized with a `KGConfig`."
            )
        logger.info(f"Initializing KG provider with config: {config}")
        self.config = config
        self.validate_config()

    def validate_config(self) -> None:
        self.config.validate()

    @property
    @abstractmethod
    def client(self) -> Any:
        """Get client."""
        pass

    @abstractmethod
    def get(self, subj: str) -> list[list[str]]:
        """Abstract method to get triplets."""
        pass

    @abstractmethod
    def get_rel_map(
        self,
        subjs: Optional[list[str]] = None,
        depth: int = 2,
        limit: int = 30,
    ) -> dict[str, list[list[str]]]:
        """Abstract method to get depth-aware rel map."""
        pass

    @abstractmethod
    def upsert_nodes(self, nodes: list[EntityNode]) -> None:
        """Abstract method to add triplet."""
        pass

    @abstractmethod
    def upsert_relations(self, relations: list[Relation]) -> None:
        """Abstract method to add triplet."""
        pass

    @abstractmethod
    def delete(self, subj: str, rel: str, obj: str) -> None:
        """Abstract method to delete triplet."""
        pass

    @abstractmethod
    def get_schema(self, refresh: bool = False) -> str:
        """Abstract method to get the schema of the graph store."""
        pass

    @abstractmethod
    def structured_query(
        self, query: str, param_map: Optional[dict[str, Any]] = {}
    ) -> Any:
        """Abstract method to query the graph store with statement and parameters."""
        pass

    @abstractmethod
    def vector_query(
        self, query: VectorStoreQuery, **kwargs: Any
    ) -> Tuple[list[LabelledNode], list[float]]:
        """Abstract method to query the graph store with a vector store query."""

    # TODO - Type this method.
    @abstractmethod
    def update_extraction_prompt(
        self,
        prompt_provider: Any,
        entity_types: list[Any],
        relations: list[Relation],
    ):
        """Abstract method to update the KG extraction prompt."""
        pass

    # TODO - Type this method.
    @abstractmethod
    def update_agent_prompt(
        self,
        prompt_provider: Any,
        entity_types: list[Any],
        relations: list[Relation],
    ):
        """Abstract method to update the KG agent prompt."""
        pass



import json
import logging
import uuid
from typing import Any, AsyncGenerator, Generator, Optional

from r2r.core import (
    AsyncState,
    GenerationConfig,
    LLMChatCompletionChunk,
    LLMProvider,
    PipeType,
    PromptProvider,
)

from .abstractions.generator_pipe import GeneratorPipe
from .search_rag_pipe import R2RSearchRAGPipe

logger = logging.getLogger(__name__)


class R2RStreamingSearchRAGPipe(R2RSearchRAGPipe):
    SEARCH_STREAM_MARKER = "search"
    COMPLETION_STREAM_MARKER = "completion"

    def __init__(
        self,
        llm_provider: LLMProvider,
        prompt_provider: PromptProvider,
        type: PipeType = PipeType.GENERATOR,
        config: Optional[GeneratorPipe] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            llm_provider=llm_provider,
            prompt_provider=prompt_provider,
            type=type,
            config=config
            or GeneratorPipe.Config(
                name="default_streaming_rag_pipe", task_prompt="default_rag"
            ),
            *args,
            **kwargs,
        )

    async def _run_logic(
        self,
        input: R2RSearchRAGPipe.Input,
        state: AsyncState,
        run_id: uuid.UUID,
        rag_generation_config: GenerationConfig,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[str, None]:
        iteration = 0
        context = ""
        # dump the search results and construct the context
        yield f"<{self.SEARCH_STREAM_MARKER}>"
        for result in input.raw_search_results:
            if iteration >= 1:
                yield ","
            yield json.dumps(result.json())
            context += f"Result {iteration+1}:\n{result.metadata['text']}\n\n"
            iteration += 1
        yield f"</{self.SEARCH_STREAM_MARKER}>"

        messages = self._get_message_payload(str(input.query), context)
        yield f"<{self.COMPLETION_STREAM_MARKER}>"
        response = ""
        for chunk in self.llm_provider.get_completion_stream(
            messages=messages, generation_config=rag_generation_config
        ):
            chunk = R2RStreamingSearchRAGPipe._process_chunk(chunk)
            response += chunk
            yield chunk

        yield f"</{self.COMPLETION_STREAM_MARKER}>"

        await self.enqueue_log(
            run_id=run_id,
            key="llm_response",
            value=response,
        )

    async def _yield_chunks(
        self,
        start_marker: str,
        chunks: Generator[str, None, None],
        end_marker: str,
    ) -> str:
        yield start_marker
        for chunk in chunks:
            yield chunk
        yield end_marker

    def _get_message_payload(
        self, query: str, context: str
    ) -> list[dict[str, str]]:
        return [
            {
                "role": "system",
                "content": self.prompt_provider.get_prompt(
                    self.config.system_prompt
                ),
            },
            {
                "role": "user",
                "content": self.prompt_provider.get_prompt(
                    self.config.task_prompt,
                    inputs={"query": query, "context": context},
                ),
            },
        ]

    @staticmethod
    def _process_chunk(chunk: LLMChatCompletionChunk) -> str:
        return chunk.choices[0].delta.content or ""



import logging
import os

import posthog

from r2r.telemetry.events import BaseTelemetryEvent

logger = logging.getLogger(__name__)


class PosthogClient:
    """
    This is a write-only project API key, so it can only create new events. It can't
    read events or any of your other data stored with PostHog, so it's safe to use in public apps.
    """

    def __init__(self, api_key: str, enabled: bool = True):
        self.enabled = enabled
        if self.enabled:
            logger.info(
                "Initializing anonymized telemetry. To disable, set TELEMETRY_ENABLED=false in your environment."
            )
            posthog.project_api_key = api_key
        else:
            posthog.disabled = True
        logger.info(
            f"Posthog telemetry {'enabled' if self.enabled else 'disabled'}"
        )

    def capture(self, event: BaseTelemetryEvent):
        if self.enabled:
            posthog.capture(event.event_id, event.event_type, event.properties)


# Initialize the telemetry client with a flag to enable or disable telemetry
telemetry_enabled = os.getenv("TELEMETRY_ENABLED", "true").lower() in (
    "true",
    "1",
    "t",
)
telemetry_client = PosthogClient(
    api_key="phc_OPBbibOIErCGc4NDLQsOrMuYFTKDmRwXX6qxnTr6zpU",
    enabled=telemetry_enabled,
)



import asyncio
import logging
import uuid
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, AsyncGenerator, Optional

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class PipeType(Enum):
    INGESTOR = "ingestor"
    EVAL = "eval"
    GENERATOR = "generator"
    SEARCH = "search"
    TRANSFORM = "transform"
    OTHER = "other"


class AsyncState:
    """A state object for storing data between pipes."""

    def __init__(self):
        self.data = {}
        self.lock = asyncio.Lock()

    async def update(self, outer_key: str, values: dict):
        """Update the state with new values."""
        async with self.lock:
            if not isinstance(values, dict):
                raise ValueError("Values must be contained in a dictionary.")
            if outer_key not in self.data:
                self.data[outer_key] = {}
            for inner_key, inner_value in values.items():
                self.data[outer_key][inner_key] = inner_value

    async def get(self, outer_key: str, inner_key: str, default=None):
        """Get a value from the state."""
        async with self.lock:
            if outer_key not in self.data:
                raise ValueError(
                    f"Key {outer_key} does not exist in the state."
                )
            if inner_key not in self.data[outer_key]:
                return default or {}
            return self.data[outer_key][inner_key]

    async def delete(self, outer_key: str, inner_key: Optional[str] = None):
        """Delete a value from the state."""
        async with self.lock:
            if outer_key in self.data and not inner_key:
                del self.data[outer_key]
            else:
                if inner_key not in self.data[outer_key]:
                    raise ValueError(
                        f"Key {inner_key} does not exist in the state."
                    )
                del self.data[outer_key][inner_key]


class AsyncPipe(ABC):
    """An asynchronous pipe for processing data."""

    class PipeConfig(BaseModel):
        """Configuration for a pipe."""

        name: str = "default_pipe"

        class Config:
            extra = "forbid"
            arbitrary_types_allowed = True

    class Input(BaseModel):
        """Input for a pipe."""

        message: AsyncGenerator[Any, None]

        class Config:
            extra = "forbid"
            arbitrary_types_allowed = True

    def __init__(
        self,
        type: PipeType = PipeType.OTHER,
        config: Optional[PipeConfig] = None,
    ):
        self._config = config or self.PipeConfig()
        self._run_info = None
        self._type = type

        logger.info(f"Initialized pipe {self.config.name} of type {self.type}")

    @property
    def config(self) -> PipeConfig:
        return self._config

    @property
    def type(self) -> PipeType:
        return self._type

    async def run(
        self,
        input: Input,
        state: AsyncState,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[Any, None]:
        result = self._run_logic(input, state)
        return result

    @abstractmethod
    async def _run_logic(
        self, input: Input, state: AsyncState, *args: Any, **kwargs: Any
    ) -> AsyncGenerator[Any, None]:
        pass



from r2r import Document, R2RAppBuilder

if __name__ == "__main__":
    r2r_app = R2RAppBuilder(from_config="neo4j_kg").build()

    r2r_app.ingest_documents(
        [
            Document(
                type="txt",
                data="John is a person that works at Google.",
                metadata={},
            ),
            Document(
                type="txt",
                data="Paul is a person that works at Microsoft that knows John.",
                metadata={},
            ),
        ]
    )

    # Get the KG provider
    neo4j_kg = r2r_app.providers.kg

    # The expected entities
    entity_names = ["John", "Paul", "Google", "Microsoft"]

    print("\nEntities:")
    for entity in entity_names:
        print(
            f"Locating {entity}:\n", neo4j_kg.get(properties={"name": entity})
        )

    relationships = neo4j_kg.get_triplets(entity_names=entity_names)

    print("\nRelationships:")
    for triplet in relationships:
        source, relation, target = triplet
        print(f"{source} -[{relation.label}]-> {target} ")

    # Search the vector database
    search_results = r2r_app.search(query="Who is john")
    print("\nSearch Results:\n", search_results)

    # Semantic search over the knowledge graph
    from r2r.core import VectorStoreQuery

    node_result = neo4j_kg.vector_query(
        VectorStoreQuery(
            query_embedding=r2r_app.providers.embedding.get_embedding(
                "A person"
            ),
        )
    )
    print("\nNode Result:", node_result)

    # Structured query
    structured_query = """
    MATCH (p1:person)-[:KNOWS]->(p2:person)
    RETURN p1.name AS Person1, p2.name AS Person2
    ORDER BY p1.name
    LIMIT 10;
    """
    print("Executing query:\n", structured_query)
    structured_result = neo4j_kg.structured_query(structured_query)
    print("Structured Results:\n", structured_result)



import asyncio
import copy
import json
import logging
import uuid
from abc import abstractmethod
from typing import Any, AsyncGenerator, Optional

from r2r.core import (
    AsyncState,
    EmbeddingProvider,
    Extraction,
    Fragment,
    FragmentType,
    KVLoggingSingleton,
    LoggableAsyncPipe,
    PipeType,
    TextSplitter,
    Vector,
    VectorEntry,
    generate_id_from_label,
)

logger = logging.getLogger(__name__)


class R2REmbeddingPipe(LoggableAsyncPipe):
    """
    Embeds and stores documents using a specified embedding model and database.
    """

    def __init__(
        self,
        embedding_provider: EmbeddingProvider,
        text_splitter: TextSplitter,
        embedding_batch_size: int = 1,
        id_prefix: str = "demo",
        pipe_logger: Optional[KVLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        config: Optional[LoggableAsyncPipe.PipeConfig] = None,
        *args,
        **kwargs,
    ):
        """
        Initializes the embedding pipe with necessary components and configurations.
        """
        super().__init__(
            pipe_logger=pipe_logger,
            type=type,
            config=config
            or LoggableAsyncPipe.PipeConfig(name="default_embedding_pipe"),
        )
        self.embedding_provider = embedding_provider
        self.text_splitter = text_splitter
        self.embedding_batch_size = embedding_batch_size
        self.id_prefix = id_prefix
        self.pipe_run_info = None

    async def fragment(
        self, extraction: Extraction, run_id: uuid.UUID
    ) -> AsyncGenerator[Fragment, None]:
        """
        Splits text into manageable chunks for embedding.
        """
        if not isinstance(extraction, Extraction):
            raise ValueError(
                f"Expected an Extraction, but received {type(extraction)}."
            )
        if not isinstance(extraction.data, str):
            raise ValueError(
                f"Expected a string, but received {type(extraction.data)}."
            )
        text_chunks = [
            ele.page_content
            for ele in self.text_splitter.create_documents([extraction.data])
        ]
        for iteration, chunk in enumerate(text_chunks):
            fragment = Fragment(
                id=generate_id_from_label(f"{extraction.id}-{iteration}"),
                type=FragmentType.TEXT,
                data=chunk,
                metadata=copy.deepcopy(extraction.metadata),
                extraction_id=extraction.id,
                document_id=extraction.document_id,
            )
            yield fragment
            fragment_dict = fragment.dict()
            await self.enqueue_log(
                run_id=run_id,
                key="fragment",
                value=json.dumps(
                    {
                        "data": fragment_dict["data"],
                        "document_id": str(fragment_dict["document_id"]),
                        "extraction_id": str(fragment_dict["extraction_id"]),
                        "fragment_id": str(fragment_dict["id"]),
                    }
                ),
            )
            iteration += 1

    async def transform_fragments(
        self, fragments: list[Fragment], metadatas: list[dict]
    ) -> AsyncGenerator[Fragment, None]:
        """
        Transforms text chunks based on their metadata, e.g., adding prefixes.
        """
        async for fragment, metadata in zip(fragments, metadatas):
            if "chunk_prefix" in metadata:
                prefix = metadata.pop("chunk_prefix")
                fragment.data = f"{prefix}\n{fragment.data}"
            yield fragment

    async def embed(self, fragments: list[Fragment]) -> list[float]:
        return await self.embedding_provider.async_get_embeddings(
            [fragment.data for fragment in fragments],
            EmbeddingProvider.PipeStage.BASE,
        )

    async def _process_batch(
        self, fragment_batch: list[Fragment]
    ) -> list[VectorEntry]:
        """
        Embeds a batch of fragments and yields vector entries.
        """
        vectors = await self.embed(fragment_batch)
        return [
            VectorEntry(
                id=fragment.id,
                vector=Vector(data=raw_vector),
                metadata={
                    "document_id": fragment.document_id,
                    "extraction_id": fragment.extraction_id,
                    "text": fragment.data,
                    **fragment.metadata,
                },
            )
            for raw_vector, fragment in zip(vectors, fragment_batch)
        ]

    async def _run_logic(
        self,
        input: LoggableAsyncPipe.Input,
        state: AsyncState,
        run_id: uuid.UUID,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[VectorEntry, None]:
        """
        Executes the embedding pipe: chunking, transforming, embedding, and storing documents.
        """
        batch_tasks = []
        fragment_batch = []

        fragment_info = {}
        async for extraction in input.message:
            async for fragment in self.fragment(extraction, run_id):
                if extraction.document_id in fragment_info:
                    fragment_info[extraction.document_id] += 1
                else:
                    fragment_info[extraction.document_id] = 1
                fragment.metadata["chunk_order"] = fragment_info[
                    extraction.document_id
                ]
                fragment_batch.append(fragment)
                if len(fragment_batch) >= self.embedding_batch_size:
                    # Here, ensure `_process_batch` is scheduled as a coroutine, not called directly
                    batch_tasks.append(
                        self._process_batch(fragment_batch.copy())
                    )  # pass a copy if necessary
                    fragment_batch.clear()  # Clear the batch for new fragments

        logger.info(
            f"Fragmented the input document ids into counts as shown: {fragment_info}"
        )

        if fragment_batch:  # Process any remaining fragments
            batch_tasks.append(self._process_batch(fragment_batch.copy()))

        # Process tasks as they complete
        for task in asyncio.as_completed(batch_tasks):
            batch_result = await task  # Wait for the next task to complete
            for vector_entry in batch_result:
                yield vector_entry



import logging
import uuid
from typing import Any, AsyncGenerator, Optional

from r2r.core import (
    AsyncPipe,
    AsyncState,
    GenerationConfig,
    LLMChatCompletion,
    LLMProvider,
    PipeType,
    PromptProvider,
    SearchResult,
)

from .abstractions.generator_pipe import GeneratorPipe

logger = logging.getLogger(__name__)


class R2RSearchRAGPipe(GeneratorPipe):
    class Input(AsyncPipe.Input):
        message: AsyncGenerator[SearchResult, None]
        query: list[str]
        raw_search_results: Optional[list[SearchResult]] = None
        dummy: Optional[str] = None

    def __init__(
        self,
        llm_provider: LLMProvider,
        prompt_provider: PromptProvider,
        type: PipeType = PipeType.GENERATOR,
        config: Optional[GeneratorPipe] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            llm_provider=llm_provider,
            prompt_provider=prompt_provider,
            type=type,
            config=config
            or GeneratorPipe.Config(
                name="default_rag_pipe", task_prompt="default_rag"
            ),
            *args,
            **kwargs,
        )

    async def _run_logic(
        self,
        input: Input,
        state: AsyncState,
        run_id: uuid.UUID,
        rag_generation_config: GenerationConfig,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[LLMChatCompletion, None]:
        context = await self._collect_context(input)
        messages = self._get_message_payload("\n".join(input.query), context)

        response = self.llm_provider.get_completion(
            messages=messages, generation_config=rag_generation_config
        )
        yield response

        await self.enqueue_log(
            run_id=run_id,
            key="llm_response",
            value=response.choices[0].message.content,
        )

    def _get_message_payload(self, query: str, context: str) -> dict:
        return [
            {
                "role": "system",
                "content": self.prompt_provider.get_prompt(
                    self.config.system_prompt,
                ),
            },
            {
                "role": "user",
                "content": self.prompt_provider.get_prompt(
                    self.config.task_prompt,
                    inputs={
                        "query": "\n".join(query),
                        "context": context,
                    },
                ),
            },
        ]

    async def _collect_context(self, input: Input) -> str:
        iteration = 0
        context = ""
        async for result in input.message:
            context += f"Result {iteration+1}:\n{result.metadata['text']}\n\n"
            iteration += 1

        return context



import logging

from .core import *
from .integrations import *
from .main import *
from .pipes import *
from .prebuilts import *
from .prompts import *

logger = logging.getLogger("r2r")
logger.setLevel(logging.INFO)

# Create a console handler and set the level to info
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)

# Create a formatter and set it for the handler
formatter = logging.Formatter(
    "%(name)s - %(levelname)s - %(message)s - %(asctime)s"
)
ch.setFormatter(formatter)

# Add the handler to the logger
logger.addHandler(ch)

# Optional: Prevent propagation to the root logger
logger.propagate = False

__all__ = [
    "LoggingConfig",
    "LocalKVLoggingProvider",
    "PostgresLoggingConfig",
    "PostgresKVLoggingProvider",
    "RedisLoggingConfig",
    "RedisKVLoggingProvider",
    "KVLoggingSingleton",
    "VectorEntry",
    "VectorType",
    "Vector",
    "SearchRequest",
    "SearchResult",
    "AsyncPipe",
    "PipeType",
    "AsyncState",
    "Prompt",
    "DataType",
    "DocumentType",
    "Document",
    "Extraction",
    "ExtractionType",
    "Fragment",
    "FragmentType",
    "SearchPipe",
    # Parsers
    "AsyncParser",
    "CSVParser",
    "DOCXParser",
    "HTMLParser",
    "JSONParser",
    "MarkdownParser",
    "PDFParser",
    "PPTParser",
    "TextParser",
    "XLSXParser",
    "Pipeline",
    # Providers
    "EmbeddingConfig",
    "EmbeddingProvider",
    "EvalConfig",
    "EvalProvider",
    "LLMEvalProvider",
    "PromptConfig",
    "PromptProvider",
    "GenerationConfig",
    "LLMChatCompletion",
    "LLMChatCompletionChunk",
    "LLMConfig",
    "LLMProvider",
    "VectorDBConfig",
    "VectorDBProvider",
    "R2RConfig",
    "TextSplitter",
    "RecursiveCharacterTextSplitter",
    "generate_run_id",
    "generate_id_from_label",
    "R2RApp",
    # Pipes
    "R2REmbeddingPipe",
    "R2REvalPipe",
    "R2RDocumentParsingPipe",
    "R2RQueryTransformPipe",
    "R2RSearchRAGPipe",
    "R2RStreamingSearchRAGPipe",
    "R2RVectorSearchPipe",
    "R2RVectorStoragePipe",
    "R2RPromptProvider",
    "R2RWebSearchPipe",
    "R2RAppBuilder",
    "R2RKGAgentPipe",
    # Prebuilts
    "MultiSearchPipe",
    "R2RPipeFactoryWithMultiSearch",
    # Integrations
    "SerperClient",
]
